import unittest

from pymcutil.constants import LONG_MIN, LONG_MAX
from pymcutil.data_tag.long_data_tag import LongDataTag


class LongDataTagTestCase(unittest.TestCase):
    def test_default(self):
        dt = LongDataTag()
        self.assertEqual(dt.value, 0)
        self.assertEqual(dt.to_str(), '0l')

    def test_positive(self):
        dt = LongDataTag(1_234)
        self.assertEqual(dt.value, 1_234)
        self.assertEqual(dt.to_str(), '1234l')

    def test_negative(self):
        dt = LongDataTag(-1_234)
        self.assertEqual(dt.value, -1_234)
        self.assertEqual(dt.to_str(), '-1234l')

    def test_float_conversion(self):
        dt = LongDataTag(1_234.5)
        self.assertEqual(dt.value, 1_234)
        self.assertEqual(dt.to_str(), '1234l')

    def test_min(self):
        dt = LongDataTag(LONG_MIN)
        self.assertEqual(dt.value, LONG_MIN)
        self.assertEqual(dt.to_str(), f'{LONG_MIN}l')

    def test_max(self):
        dt = LongDataTag(LONG_MAX)
        self.assertEqual(dt.value, LONG_MAX)
        self.assertEqual(dt.to_str(), f'{LONG_MAX}l')

    def test_too_low(self):
        with self.assertRaises(ValueError):
            LongDataTag(LONG_MIN - 1)

    def test_too_high(self):
        with self.assertRaises(ValueError):
            LongDataTag(LONG_MAX + 1)

    def test_equality(self):
        dt1 = LongDataTag(1_234)
        dt2 = LongDataTag(1_234)
        self.assertEqual(dt1, dt2)



from .._execute_conditional_entity_mixin import ExecuteConditionalEntityMixin
from .._execute_unless_command import ExecuteUnlessCommand

CMD = 'entity'


class ExecuteUnlessEntityCommand(ExecuteUnlessCommand, ExecuteConditionalEntityMixin):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute unless entity <condition: entities> -> execute

    http://minecraft.gamepedia.com/Commands#execute
    """

    def params(self):
        yield from super().params()
        yield from (
            CMD, self.target,
            (self.subcommand.substr(), None))


entity = ExecuteUnlessEntityCommand



import abc


class ResourceManager(abc.ABC):
    @abc.abstractmethod
    def register(self, kind, generator, database=None): ...

    @abc.abstractmethod
    def get(self, reference): ...

    @abc.abstractmethod
    def generate(self, reference): ...

    @abc.abstractmethod
    def locate(self, reference): ...

    @abc.abstractmethod
    def name(self, reference): ...

    @abc.abstractmethod
    def path(self, reference): ...



import unittest

from pymcutil.command import commands


class SayCommandTestCase(unittest.TestCase):
    def test_without_spaces(self):
        cmd = commands.say(
            message='hi')
        self.assertEqual(str(cmd), 'say hi')

    def test_with_spaces(self):
        cmd = commands.say(
            message='hello world')
        self.assertEqual(str(cmd), 'say hello world')



import abc

from pymcutil.resource.abc.resource import Resource
from pymcutil.resource.resource_location.abc.resource_location import ResourceLocation
from pymcutil.resource.resource_manager.abc.resource_manager import ResourceManager
from pymcutil.resource.resource_reference.abc.resource_reference import ResourceReference


class ResourceGenerator(abc.ABC):
    @abc.abstractmethod
    def location(self, manager: ResourceManager, reference: ResourceReference) -> ResourceLocation:
        """ Return a `ResourceLocation` for the resource that would be generated by the given `ResourceReference`. """

    @abc.abstractmethod
    def generate(self, manager: ResourceManager, reference: ResourceReference) -> Resource:
        """ Return a `Resource` generated using the given `ResourceReference`."""



from pymcutil.data_tag import DataTag


class FloatDataTag(DataTag):
    """ A float data tag, also known as TAG_Float. """

    def __init__(self, value=0.0):
        self._value: float = None
        self.value = value

    def __eq__(self, other):
        return isinstance(other, FloatDataTag) and self.value.__eq__(other.value)

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value):
        self._value = float(value)

    def to_str(self) -> str:
        return f'{self.value}f'



from pymcutil.command.command import Command

CMD = 'execute'


class ExecuteCommand(Command):
    """
    An objective model of Minecraft's `execute` command:

        execute ...

    http://minecraft.gamepedia.com/Commands#execute

    *This command requires a sub-command to be useful.*
    """

    def params(self):
        yield CMD



from pymcutil.util.symbol import Symbol


class SelectorSort(Symbol):
    pass


nearest = SelectorSort('nearest')
furthest = SelectorSort('furthest')
random = SelectorSort('random')
arbitrary = SelectorSort('arbitrary')



import abc
from typing import Any


class Serializable(abc.ABC):
    """ An object that can be serialized into a composition of basic values. """

    @property
    @abc.abstractmethod
    def serialized(self) -> Any:
        """ Returns a serialized representation of the object. This *serialized representation* should be composed
        entirely of atomic values (ints, strs, bools, etc), lists, and dicts. Simply put, it should be serializable into
        JSON using `json.dump()`. """



from pymcutil.command.command import Command

CMD = 'say'


class SayCommand(Command):
    """
    An objective model of Minecraft's `say` command:

        say <message ...>

    http://minecraft.gamepedia.com/Commands#say
    """

    def __init__(self, message: str):
        self.message: str = message

    def params(self):
        yield from (CMD, self.message)


say = SayCommand



from pymcutil.selector.abc.selector import Selector
from pymcutil.symbols import selector_bases
from pymcutil.symbols.selector_bases.selector_bases import SelectorBase


class EntitiesSelector(Selector):
    @property
    def base(self) -> SelectorBase:
        return selector_bases.entities


entities = EntitiesSelector

ENTITIES = EntitiesSelector()



import logging
import tempfile
import unittest

from pymcutil.resource.function_resource import FunctionResource
from pymcutil.resource.resource_database.abc.resource_database import ResourceDatabase
from pymcutil.resource.resource_database.forgetful_resource_database import ForgetfulResourceDatabase
from pymcutil.resource.resource_manager.abc.resource_manager import ResourceManager
from pymcutil.resource.resource_manager.standard_resource_manager import StandardResourceManager
from . import functions, procedures

log = logging.getLogger(__name__)


class StringSplitE2ETestCase(unittest.TestCase):
    def test(self):
        tempdir = tempfile.TemporaryDirectory()
        log.info('Using temporary directory: {}'.format(tempdir.name))

        database: ResourceDatabase = ForgetfulResourceDatabase(tempdir.name)

        registry = (
            (functions.string_split, procedures.string_split(namespace='string_split')),
        )

        manager: ResourceManager = StandardResourceManager(database, registry)

        root_lrs = list(manager.generate(reference) for reference in (
            functions.string_split(message='abcdefghi', indent=0),
            functions.string_split(message='xyz', indent=0),
        ))

        log.info('Generated {} root resources:'.format(len(root_lrs)))
        for root_lr in root_lrs:
            log.info('    [{}] {} ({})'.format(
                type(root_lr.resource).__name__, root_lr.location.name, root_lr.location.path))

        all_lrs = list(database.all())
        log.info('Generated {} resources overall:'.format(len(all_lrs)))
        for lr in all_lrs:
            location, resource = lr.location, lr.resource
            log.info('    [{}] {} ({})'.format(
                type(resource).__name__, location.name, location.path))
            if isinstance(resource, FunctionResource):
                for line in resource.lines:
                    log.debug('        {}'.format(line))

        tempdir.cleanup()



from pymcutil.position.rc import RC
from pymcutil.position.vector_2d import Vector2D


class Rotation(Vector2D):
    """ Rotation representation using yaw and pitch. Includes relative capabilities. """

    def __init__(self, x: float, y: float):
        super().__init__(x, y)

    def __invert__(self):
        return self.relative()

    def __abs__(self):
        return self.absolute()

    def relative(self):
        return self.__class__(*[RC(c) for c in self.components])

    def absolute(self):
        return self.__class__(*[float(c) for c in self.components])

    @property
    def yaw(self):
        return self.x

    @property
    def pitch(self):
        return self.y


ZERO_ROTATION = Rotation(0, 0)
ZERO_TURN = ~ZERO_ROTATION



import abc
from typing import Iterable, Union

from pymcutil.resource.abc.resource import Resource
from pymcutil.resource.located_resource.located_resource import LocatedResource
from pymcutil.resource.resource_location.abc.resource_location import ResourceLocation


class ResourceDatabase(abc.ABC):
    @abc.abstractmethod
    def get(self, location: ResourceLocation) -> Union[LocatedResource, None]:
        """ Return the `LocatedResource` at the given `location` if available, otherwise `None`. """

    @abc.abstractmethod
    def put(self, location: ResourceLocation, resource: Resource) -> LocatedResource:
        """ Persist and return a `LocatedResource` consisting of the given `resource` at the given `location`. """

    @abc.abstractmethod
    def all(self) -> Iterable[LocatedResource]:
        """ Iterate through all available `LocatedResource`s. """



import collections
from typing import Union

from pymcutil.util.siftable import Siftable


class Symbol(Siftable, collections.Hashable):
    Generic = Union['Symbol', str]

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return isinstance(other, Symbol) and \
               isinstance(other, self.__class__) \
               and (other.value == self.value)

    @classmethod
    def _from_other(cls, obj):
        if isinstance(obj, str):
            return cls(obj)



import typing

from pymcutil.resource.resource_location.abc.resource_location import ResourceLocation, ResourceName, ResourcePath


class StandardResourceLocation(ResourceLocation):
    def __init__(self, namespace: str, components: typing.Iterable[str], subfolder: str, extension: str = 'json'):
        self._namespace: str = namespace
        self._components: typing.Tuple[str] = tuple(components)  # save a copy
        self._subfolder: str = subfolder
        self._extension: str = extension

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        return isinstance(other, StandardResourceLocation) \
               and isinstance(other, self.__class__) \
               and (self.path == other.path)

    @property
    def namespace(self) -> str:
        return self._namespace

    @property
    def components(self) -> typing.Iterable[str]:
        yield from self._components

    @property
    def trail(self) -> str:
        # TODO use utility
        return '/'.join(self.components)

    @property
    def subfolder(self) -> str:
        return self._subfolder

    @property
    def extension(self) -> str:
        return self._extension

    @property
    def name(self) -> ResourceName:
        # TODO use utility
        return '{}:{}'.format(self.namespace, self.trail)

    @property
    def path(self) -> ResourcePath:
        # TODO use utility with os-specific separator
        return 'data/{}/{}/{}.{}'.format(self.namespace, self.subfolder, self.trail, self.extension)



import os
import unittest

from pymcutil.util.resources import get_advancement_path, get_function_path, get_loot_table_path, get_resource_name


class ResourcesTestCase(unittest.TestCase):
    def test_get_resource_name(self):
        self.assertEqual(
            get_resource_name(namespace='stuff', components=('foo', 'bar')),
            'stuff:foo/bar')

    def test_get_advancement_path(self):
        self.assertEqual(
            get_advancement_path(root='.', namespace='stuff', components=('foo', 'bar')),
            '{}.{}'.format(os.sep.join(('.', 'data', 'stuff', 'advancements', 'foo', 'bar')), 'json'))

    def test_get_function_path(self):
        self.assertEqual(
            get_function_path(root='.', namespace='stuff', components=('foo', 'bar')),
            '{}.{}'.format(os.sep.join(('.', 'data', 'stuff', 'functions', 'foo', 'bar')), 'mcfunction'))

    def test_get_loot_table_path(self):
        self.assertEqual(
            get_loot_table_path(root='.', namespace='stuff', components=('foo', 'bar')),
            '{}.{}'.format(os.sep.join(('.', 'data', 'stuff', 'loot_tables', 'foo', 'bar')), 'json'))



import unittest

from pymcutil.block.block import Block
from pymcutil.block.block_state import BlockState


class BlockTestCase(unittest.TestCase):
    def test_block_with_id(self):
        b = Block('piston')
        self.assertEqual(b.block_id, 'piston')

    def test_block_with_id_and_block_state(self):
        b = Block('piston', BlockState(facing='up'))
        self.assertEqual(b.block_id, 'piston')
        self.assertEqual(b.block_state, BlockState(facing='up'))

    def test_block_state_from_dict(self):
        b = Block('piston', dict(facing='up'))
        self.assertEqual(b.block_state, BlockState(facing='up'))

    def test_block_state_from_braces(self):
        b = Block('piston', {'facing': 'up'})
        self.assertEqual(b.block_state, BlockState(facing='up'))

    def test_state(self):
        b1 = Block('piston')
        b2 = b1.state(facing='up')
        self.assertEqual(b1.block_id, 'piston')
        self.assertEqual(b1.block_state, BlockState())
        self.assertEqual(b2.block_id, 'piston')
        self.assertEqual(b2.block_state, BlockState(facing='up'))

    def test_state_overwrite(self):
        b1 = Block('piston', dict(facing='up'))
        b2 = b1.state(facing='down')
        self.assertEqual(b1.block_id, 'piston')
        self.assertEqual(b1.block_state, BlockState(facing='up'))
        self.assertEqual(b2.block_id, 'piston')
        self.assertEqual(b2.block_state, BlockState(facing='down'))

    def test_state_more(self):
        b1 = Block('piston', dict(facing='up'))
        b2 = b1.state(extended=True)
        self.assertEqual(b1.block_id, 'piston')
        self.assertEqual(b1.block_state, BlockState(facing='up'))
        self.assertEqual(b2.block_id, 'piston')
        self.assertEqual(b2.block_state, BlockState(facing='up', extended=True))



from collections import MutableSequence
from typing import Iterable


class Stack(MutableSequence):
    def __init__(self, items: Iterable = ()):
        self._items = list(items)

    def __getitem__(self, index):
        return self._items.__getitem__(index)

    def __setitem__(self, index, value):
        return self._items.__setitem__(index, value)

    def __delitem__(self, index):
        return self._items.__delitem__(index)

    def __len__(self):
        return self._items.__len__()

    def insert(self, index, value):
        return self._items.insert(index, value)

    def push(self, value):
        self._items.append(value)

    def empty(self) -> bool:
        return not bool(self)



import collections
import typing

from pymcutil.util.siftable import MappingSiftable


class CommonMapping(collections.MutableMapping, MappingSiftable):
    Generic = typing.Mapping

    def __init__(self, **kwargs):
        self._mapping = {}
        self.update(kwargs)

    def __setitem__(self, key, value):
        return self._mapping.__setitem__(key, value)

    def __delitem__(self, key):
        return self._mapping.__delitem__(key)

    def __getitem__(self, key):
        return self._mapping.__getitem__(key)

    def __len__(self):
        return self._mapping.__len__()

    def __iter__(self):
        return self._mapping.__iter__()



import abc
from typing import Iterable

from pymcutil.command.command import Command
from pymcutil.resource.function_resource import FunctionResource
from pymcutil.resource.resource_generator.abc.function_resource_generator import FunctionResourceGenerator
from pymcutil.resource.resource_location.function_resource_location import FunctionResourceLocation
from pymcutil.resource.resource_manager.abc.resource_manager import ResourceManager
from pymcutil.resource.resource_reference.function_resource_reference import FunctionResourceReference


class Procedure(FunctionResourceGenerator):
    def __init__(self, namespace: str, root: str = None):
        self.namespace: str = namespace
        self.root: str = root

    def location(self, manager: ResourceManager, reference: FunctionResourceReference):
        return FunctionResourceLocation(namespace=self.namespace, components=self.components(manager, reference))

    def generate(self, manager: ResourceManager, reference: FunctionResourceReference):
        return FunctionResource(commands=tuple(self.commands(manager, reference)))

    def components(self, manager: ResourceManager, reference: FunctionResourceReference) -> Iterable[str]:
        """ Yield components making up the trailing portion of the function name. """
        yield from (str(param) for param in reference.params)

    @abc.abstractmethod
    def commands(self, manager: ResourceManager, reference: FunctionResourceReference) -> Iterable[Command]:
        """ Yield commands, in order, making up the function. """



from collections import Mapping

from pymcutil.block.block import Block
from pymcutil.data_tag.compound_data_tag import CompoundDataTag


class BlockEntity(Block):
    """ Represents a Minecraft block entity. """

    def __init__(self, block_id: str, block_state: Mapping = None, data_tag: Mapping = None):
        super().__init__(block_id=block_id, block_state=block_state)
        self.data_tag: CompoundDataTag = CompoundDataTag.sift(data_tag, {})

    def __str__(self):
        return ''.join(
            (super().__str__(),
             str(self.data_tag) if self.data_tag else ''))

    def tag(self, *args, **params):
        new_data_tag = dict(self.data_tag)  # copy
        new_data_tag.update(*args, **params)
        return BlockEntity(block_id=self.block_id, block_state=self.block_state, data_tag=new_data_tag)



import abc
from typing import Iterator, Any


class Command(abc.ABC):
    """ An abstract base class representing a Minecraft command. """

    def __str__(self):
        return self.to_str()

    @staticmethod
    def convert_param(param) -> str:
        if isinstance(param, bool):
            return 'true' if param else 'false'
        return str(param)

    def to_str(self) -> str:
        # TODO Clean up this mess.
        params = []
        skip = True
        for p in reversed(list(self.params())):
            value, default = p if isinstance(p, tuple) else (p, None)
            if skip:
                skip = value is None and default is not None
                if skip:
                    continue
            final_value = value if value is not None else default
            if final_value is None:
                raise ValueError('Command parameter required but not given')
            params.append(final_value)
        return ' '.join(self.convert_param(p) for p in reversed(params))

    def substr(self) -> str:
        """ A little hack to support non-prefixed subcommands, such as `execute`. """
        # TODO There's got to be something cleaner.
        return self.to_str().split(maxsplit=1)[1]

    @abc.abstractmethod
    def params(self) -> Iterator[Any]:
        """ Yield command parameters as pairs of (value, default). """



from pymcutil.position.vector import Vector


class Vector3D(Vector):
    def __init__(self, x: float, y: float, z: float):
        super().__init__(x, y, z)

    @property
    def x(self) -> float:
        return self._components[0]

    @property
    def y(self) -> float:
        return self._components[1]

    @property
    def z(self) -> float:
        return self._components[2]



from .._execute_unless_blocks_command import ExecuteUnlessBlocksCommand

CMD = 'all'


class ExecuteUnlessBlocksAllCommand(ExecuteUnlessBlocksCommand):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute unless blocks <start: x y z> <end: x y z> <destination: x y z> all -> execute

    http://minecraft.gamepedia.com/Commands#execute
    """

    def params(self):
        yield from super().params()
        yield from (
            CMD,
            (self.subcommand.substr(), None))


all = ExecuteUnlessBlocksAllCommand



import json

from pymcutil.resource.abc.resource import Resource
from pymcutil.util.serializable import Serializable


class JsonResource(Resource, Serializable):
    def write_to(self, path):
        with open(path, 'w') as fp:
            json.dump(self.serialized, fp, sort_keys=True, indent=2)



import unittest

from pymcutil.data_tag.double_data_tag import DoubleDataTag


class DoubleDataTagTestCase(unittest.TestCase):
    def test_default(self):
        dt = DoubleDataTag()
        self.assertEqual(dt.value, 0.0)
        self.assertEqual(dt.to_str(), '0.0d')

    def test_positive(self):
        dt = DoubleDataTag(456.5)
        self.assertEqual(dt.value, 456.5)
        self.assertEqual(dt.to_str(), '456.5d')

    def test_negative(self):
        dt = DoubleDataTag(-456.5)
        self.assertEqual(dt.value, -456.5)
        self.assertEqual(dt.to_str(), '-456.5d')

    def test_int_conversion(self):
        dt = DoubleDataTag(456)
        self.assertEqual(dt.value, 456.0)
        self.assertEqual(dt.to_str(), '456.0d')

    def test_equality(self):
        dt1 = DoubleDataTag(456.5)
        dt2 = DoubleDataTag(456.5)
        self.assertEqual(dt1, dt2)



import unittest

from pymcutil.constants import SHORT_MIN, SHORT_MAX
from pymcutil.data_tag.short_data_tag import ShortDataTag


class ShortDataTagTestCase(unittest.TestCase):
    def test_default(self):
        dt = ShortDataTag()
        self.assertEqual(dt.value, 0)
        self.assertEqual(dt.to_str(), '0s')

    def test_positive(self):
        dt = ShortDataTag(1_234)
        self.assertEqual(dt.value, 1_234)
        self.assertEqual(dt.to_str(), '1234s')

    def test_negative(self):
        dt = ShortDataTag(-1_234)
        self.assertEqual(dt.value, -1_234)
        self.assertEqual(dt.to_str(), '-1234s')

    def test_float_conversion(self):
        dt = ShortDataTag(1_234.5)
        self.assertEqual(dt.value, 1_234)
        self.assertEqual(dt.to_str(), '1234s')

    def test_min(self):
        dt = ShortDataTag(SHORT_MIN)
        self.assertEqual(dt.value, SHORT_MIN)
        self.assertEqual(dt.to_str(), f'{SHORT_MIN}s')

    def test_max(self):
        dt = ShortDataTag(SHORT_MAX)
        self.assertEqual(dt.value, SHORT_MAX)
        self.assertEqual(dt.to_str(), f'{SHORT_MAX}s')

    def test_too_low(self):
        with self.assertRaises(ValueError):
            ShortDataTag(SHORT_MIN - 1)

    def test_too_high(self):
        with self.assertRaises(ValueError):
            ShortDataTag(SHORT_MAX + 1)

    def test_equality(self):
        dt1 = ShortDataTag(1_234)
        dt2 = ShortDataTag(1_234)
        self.assertEqual(dt1, dt2)



import unittest

from pymcutil.position.position import Position, ZERO_POSITION, ZERO_OFFSET
from pymcutil.position.rc import rel


class PositionTestCase(unittest.TestCase):
    def test_init(self):
        p = Position(1, 2, 3)
        self.assertEqual(p.x, 1)
        self.assertEqual(p.y, 2)
        self.assertEqual(p.z, 3)

    def test_invert(self):
        self.assertEqual(str(~Position(1, 2, 3)), '~1 ~2 ~3')

    def test_abs(self):
        self.assertEqual(str(abs(~Position(1, 2, 3))), '1 2 3')

    def test_zero_position(self):
        self.assertEqual(str(ZERO_POSITION), '0 0 0')

    def test_zero_offset(self):
        self.assertEqual(str(ZERO_OFFSET), '~ ~ ~')

    def test_rel_mixed_zero_coords(self):
        self.assertEqual(str(~Position(4, 0, 4)), '~4 ~ ~4')

    def test_rel_decimal_coords(self):
        self.assertEqual(str(~Position(3.142, 6.674, 2.718)), '~3.142 ~6.674 ~2.718')

    def test_mixed_abs_and_rel_coords(self):
        self.assertEqual(str(Position(1, rel(2), 3)), '1 ~2 3')

    def test_mixed_abs_and_rel_to_rel_coords(self):
        self.assertEqual(str(~Position(1, rel(2), 3)), '~1 ~2 ~3')



from pymcutil.entity.abc.entity import Entity
from pymcutil.symbols import entity_ids
from pymcutil.symbols.entity_ids.entity_ids import EntityID


class AreaEffectCloud(Entity):
    def __init__(
            self,
            age: int = None, color: int = None, duration: int = None, reapplication_delay: int = None,
            wait_time: int = None, owner_uuid_least: int = None, owner_uuid_most: int = None,
            duration_on_use: float = None, radius: float = None, radius_on_use: float = None,
            radius_per_tick: float = None, particle: str = None,
            particle_param_1: int = None, particle_param_2: int = None, potion: str = None,
            effects: list = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        # TODO area effect cloud attribs

    @property
    def id_(self) -> EntityID:
        return entity_ids.area_effect_cloud



import re
import typing

from pymcutil.util.argument_mapping import ArgumentMapping


class BlockState(ArgumentMapping):
    """ Represents a Minecraft block state. """

    Generic = typing.Mapping

    PATTERN = re.compile('^[a-z0-9_]+$')

    def __setitem__(self, key, value):
        if not (isinstance(key, (bool, int)) or (isinstance(key, str) and self.PATTERN.match(key))):
            raise KeyError('Invalid block state key: {}'.format(key))

        if not (isinstance(value, (bool, int)) or (isinstance(value, str) and self.PATTERN.match(value))):
            raise ValueError('Invalid block state value: {}'.format(value))

        return super().__setitem__(key, value)

    def __str__(self):
        return 'default' if not self else self.innards



import abc
from typing import List

from pymcutil.position.position import Position
from pymcutil.position.rotation import Rotation
from pymcutil.symbols.entity_ids.entity_ids import EntityID


class Entity(abc.ABC):
    def __init__(
            self,
            pos: Position.Generic = None, motion: Position.Generic = None, rotation: Rotation.Generic = None,
            fall_distance: float = None, fire: int = None, air: int = None, on_ground: bool = None,
            no_gravity: bool = None, dimension: int = None, invulnerable: bool = None, portal_cooldown: int = None,
            uuid_most: int = None, uuid_least: int = None, custom_name: str = None, custom_name_visible: bool = None,
            silent: bool = None, passengers: List[object] = None, glowing: bool = None, tags: List[str] = None,
            **kwargs
    ):
        ...  # TODO entity attribs

    @property
    @abc.abstractmethod
    def id_(self) -> EntityID:
        """ Return the namespaced `EntityID` for this type of entity. """



import os
from typing import Iterable

from pymcutil.constants.resources import *


def get_resource_trail(*components) -> str:
    return '/'.join(str(c) for c in components if c is not None)


def get_resource_name(namespace: str, components: Iterable) -> str:
    return '{}:{}'.format(namespace, get_resource_trail(*components))


def get_resource_path(
        root: str, resource_dir: str, namespace: str, resource_subdir: str, resource_ext: str,
        components: Iterable) -> str:
    return '{}.{}'.format(os.path.join(root, resource_dir, namespace, resource_subdir, *components), resource_ext)


def get_advancement_path(root: str, namespace: str, components: Iterable) -> str:
    return get_resource_path(
        root=root, resource_dir=DATA_DIRECTORY, namespace=namespace, resource_subdir=ADVANCEMENTS_DIRECTORY,
        resource_ext=ADVANCEMENTS_EXTENSION, components=components)


def get_function_path(root: str, namespace: str, components: Iterable) -> str:
    return get_resource_path(
        root=root, resource_dir=DATA_DIRECTORY, namespace=namespace, resource_subdir=FUNCTIONS_DIRECTORY,
        resource_ext=FUNCTIONS_EXTENSION, components=components)


def get_loot_table_path(root: str, namespace: str, components: Iterable) -> str:
    return get_resource_path(
        root=root, resource_dir=DATA_DIRECTORY, namespace=namespace, resource_subdir=LOOT_TABLES_DIRECTORY,
        resource_ext=LOOT_TABLES_EXTENSION, components=components)



from .._execute_conditional_block_mixin import ExecuteConditionalBlockMixin
from .._execute_unless_command import ExecuteUnlessCommand

CMD = 'block'


class ExecuteUnlessBlockCommand(ExecuteUnlessCommand, ExecuteConditionalBlockMixin):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute unless block <pos: x y z> <block: block> -> execute

    http://minecraft.gamepedia.com/Commands#execute
    """

    def params(self):
        yield from super().params()
        yield from (
            CMD, self.position, self.block,
            (self.subcommand.substr(), None))


block = ExecuteUnlessBlockCommand



from pymcutil.resource.abc.resource import Resource
from pymcutil.resource.resource_location.abc.resource_location import ResourceLocation


class LocatedResource(object):
    """ Uses components to pair a `ResourceLocation` with a `Resource`. """

    def __init__(self, location: ResourceLocation, resource: Resource):
        self.location: ResourceLocation = location
        self.resource: Resource = resource



import unittest

from pymcutil.constants import INT_MIN, INT_MAX
from pymcutil.data_tag.int_data_tag import IntDataTag


class IntDataTagTestCase(unittest.TestCase):
    def test_default(self):
        dt = IntDataTag()
        self.assertEqual(dt.value, 0)
        self.assertEqual(dt.to_str(), '0')

    def test_positive(self):
        dt = IntDataTag(123_456_789)
        self.assertEqual(dt.value, 123_456_789)
        self.assertEqual(dt.to_str(), '123456789')

    def test_negative(self):
        dt = IntDataTag(-123_456_789)
        self.assertEqual(dt.value, -123_456_789)
        self.assertEqual(dt.to_str(), '-123456789')

    def test_float_conversion(self):
        dt = IntDataTag(123_456_789.5)
        self.assertEqual(dt.value, 123_456_789)
        self.assertEqual(dt.to_str(), '123456789')

    def test_min(self):
        dt = IntDataTag(INT_MIN)
        self.assertEqual(dt.value, INT_MIN)
        self.assertEqual(dt.to_str(), f'{INT_MIN}')

    def test_max(self):
        dt = IntDataTag(INT_MAX)
        self.assertEqual(dt.value, INT_MAX)
        self.assertEqual(dt.to_str(), f'{INT_MAX}')

    def test_too_low(self):
        with self.assertRaises(ValueError):
            IntDataTag(INT_MIN - 1)

    def test_too_high(self):
        with self.assertRaises(ValueError):
            IntDataTag(INT_MAX + 1)

    def test_equality(self):
        dt1 = IntDataTag(123_456_789)
        dt2 = IntDataTag(123_456_789)
        self.assertEqual(dt1, dt2)



import abc

from pymcutil.resource.function_resource import FunctionResource
from pymcutil.resource.resource_generator.abc.resource_generator import ResourceGenerator
from pymcutil.resource.resource_location.function_resource_location import FunctionResourceLocation
from pymcutil.resource.resource_manager.abc.resource_manager import ResourceManager
from pymcutil.resource.resource_reference.function_resource_reference import FunctionResourceReference


class FunctionResourceGenerator(ResourceGenerator):
    @abc.abstractmethod
    def location(self, manager: ResourceManager, reference: FunctionResourceReference) -> FunctionResourceLocation: ...

    @abc.abstractmethod
    def generate(self, manager: ResourceManager, reference: FunctionResourceReference) -> FunctionResource: ...



import abc

from pymcutil.data_tag import DataTag
from pymcutil.resource.abc.resource import Resource


class NBTResource(Resource):
    def write_to(self, path):
        with open(path, 'wb') as fp:
            fp.write(self.nbt.to_bytes())

    @property
    @abc.abstractmethod
    def nbt(self) -> DataTag:
        """ Return the root tag of the NBT representation of the object. """



import unittest

from pymcutil.data_tag.byte_data_tag import ByteDataTag
from pymcutil.data_tag.compound_data_tag import CompoundDataTag
from pymcutil.data_tag.double_data_tag import DoubleDataTag
from pymcutil.data_tag.helper import data_taggify
from pymcutil.data_tag.int_data_tag import IntDataTag
from pymcutil.data_tag.list_data_tag import ListDataTag
from pymcutil.data_tag.string_data_tag import StringDataTag


class HelperTestCase(unittest.TestCase):
    def test_convert_data_tag(self):
        dt1 = data_taggify('hello')
        dt2 = data_taggify(dt1)
        self.assertEqual(dt1, dt2)

    def test_convert_bool(self):
        self.assertEqual(data_taggify(True), ByteDataTag(True))

    def test_convert_int(self):
        self.assertEqual(data_taggify(123456789), IntDataTag(123456789))

    def test_convert_double(self):
        self.assertEqual(data_taggify(123.5), DoubleDataTag(123.5))

    def test_convert_str(self):
        self.assertEqual(data_taggify('hello'), StringDataTag('hello'))

    def test_convert_list(self):
        self.assertEqual(data_taggify([]), ListDataTag([]))

    def test_convert_dict(self):
        self.assertEqual(data_taggify({}), CompoundDataTag({}))

    def test_unconvertible(self):
        with self.assertRaises(TypeError):
            data_taggify(None)



from pymcutil.command.command import Command

CMD = 'effect'


class EffectCommand(Command):
    """
    An objective model of Minecraft's `effect` command:

        effect <give|take> ...

    http://minecraft.gamepedia.com/Commands#effect

    *This command requires a sub-command to be useful.*
    """

    def params(self):
        yield CMD



import unittest

from pymcutil.command import commands
from pymcutil.selector import selectors


class ScoreboardPlayersTagAddCommandTestCase(unittest.TestCase):
    def test(self):
        cmd = commands.scoreboard.players.tag.list(
            target=selectors.SELF)
        self.assertEqual(str(cmd), 'scoreboard players tag @s list')



import typing

from pymcutil.resource.resource_location.tag_resource_location import TagResourceLocation


class ItemTagResourceLocation(TagResourceLocation):
    def __init__(self, namespace: str, components: typing.Iterable[str]):
        super().__init__(namespace=namespace, components=('items', *components))



from pymcutil.constants import SHORT_MIN, SHORT_MAX
from pymcutil.data_tag import DataTag


class ShortDataTag(DataTag):
    """ A short data tag, also known as TAG_Short. """

    def __init__(self, value=0):
        self._value: int = None
        self.value = value

    def __eq__(self, other):
        return isinstance(other, ShortDataTag) and self.value.__eq__(other.value)

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, value):
        value = int(value)
        if not (SHORT_MIN <= value <= SHORT_MAX):
            raise ValueError(f'Value exceeds capacity of a short: {value}')
        self._value = value

    def to_str(self) -> str:
        return f'{self.value}s'



import unittest

from pymcutil.selector import selectors


class SelectorTestCase(unittest.TestCase):
    def test_base_player(self):
        self.assertEqual(
            '@p',
            str(selectors.PLAYER))

    def test_base_random(self):
        self.assertEqual(
            '@r',
            str(selectors.RANDOM))

    def test_base_all_players(self):
        self.assertEqual(
            '@a',
            str(selectors.ALL_PLAYERS))

    def test_base_entities(self):
        self.assertEqual(
            '@e',
            str(selectors.ENTITIES))

    def test_base_self(self):
        self.assertEqual(
            '@s',
            str(selectors.SELF))

    def test_argument_position(self):
        self.assertEqual(
            '@e[x=1.0,y=2.0,z=3.0]',
            str(selectors.entities(position=(1, 2, 3))))

    def test_argument_position_with_floats(self):
        self.assertEqual(
            '@e[x=1.9,y=2.5,z=3.1]',
            str(selectors.entities(position=(1.9, 2.5, 3.1))))

    def test_argument_volume(self):
        self.assertEqual(
            '@e[dx=4.0,dy=9.0,dz=16.0]',
            str(selectors.entities(volume=(4, 9, 16))))

    def test_argument_volume_with_floats(self):
        self.assertEqual(
            '@e[dx=4.9,dy=9.5,dz=16.1]',
            str(selectors.entities(volume=(4.9, 9.5, 16.1))))

    def test_argument_tags(self):
        self.assertEqual(
            '@e[tag=foo,tag=bar]',
            str(selectors.entities(tags=['foo', 'bar'])))

    def test_argument_not_tags(self):
        self.assertEqual(
            '@e[tag=!foo,tag=!bar]',
            str(selectors.entities(not_tags=['foo', 'bar'])))

    # def test_argument_type(self):
    #     self.assertEqual(
    #         '@e[type=area_effect_cloud]',
    #         str(selectors.entities(type='area_effect_cloud')))
    #
    # def test_argument_not_types(self):
    #     self.assertEqual(
    #         '@e[type=!area_effect_cloud]',
    #         str(selectors.entities(not_types=['area_effect_cloud'])))
    #
    # def test_argument_not_entity_types(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_type=('cow', 'pig'))),
    #         '@e[type=!cow,type=!pig]')
    #
    # def test_argument_l(self):
    #     self.assertEqual(
    #         str(selectors.entities(l=10)),
    #         '@e[l=10]')
    #
    # def test_argument_lm(self):
    #     self.assertEqual(
    #         str(selectors.entities(lm=5)),
    #         '@e[lm=5]')
    #
    # def test_argument_m(self):
    #     self.assertEqual(
    #         str(selectors.entities(m='survival')),
    #         '@e[m=survival]')
    #
    # def test_argument_not_m(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_m='survival')),
    #         '@e[m=!survival]')
    #
    # def test_argument_not_ms(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_m=('survival', 'adventure'))),
    #         '@e[m=!survival,m=!adventure]')
    #
    # def test_argument_team(self):
    #     self.assertEqual(
    #         str(selectors.entities(team='myteam')),
    #         '@e[team=myteam]')
    #
    # def test_argument_not_team(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_team='myteam')),
    #         '@e[team=!myteam]')
    #
    # def test_argument_not_teams(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_team=('myteam', 'otherteam'))),
    #         '@e[team=!myteam,team=!otherteam]')
    #
    # def test_argument_max_scores(self):
    #     self.assertEqual(
    #         str(selectors.entities(max_scores={'myobjective': 100})),
    #         '@e[score_myobjective=100]')
    #
    # def test_argument_min_scores(self):
    #     self.assertEqual(
    #         str(selectors.entities(min_scores={'myobjective': 1})),
    #         '@e[score_myobjective_min=1]')
    #
    # def test_argument_exact_score(self):
    #     self.assertEqual(
    #         str(selectors.entities(exact_scores={'myobjective': 50})),
    #         '@e[score_myobjective=50,score_myobjective_min=50]')
    #
    # def test_argument_name(self):
    #     self.assertEqual(
    #         str(selectors.entities(name='myname')),
    #         '@e[name=myname]')
    #
    # def test_argument_not_name(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_name='myname')),
    #         '@e[name=!myname]')
    #
    # def test_argument_not_names(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_name=('myname', 'othername'))),
    #         '@e[name=!myname,name=!othername]')
    #
    # def test_argument_tag(self):
    #     self.assertEqual(
    #         str(selectors.entities(tag='mytag')),
    #         '@e[tag=mytag]')
    #
    # def test_argument_tags(self):
    #     self.assertEqual(
    #         str(selectors.entities(tag=('mytag', 'othertag'))),
    #         '@e[tag=mytag,tag=othertag]')
    #
    # def test_argument_not_tag(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_tag='mytag')),
    #         '@e[tag=!mytag]')
    #
    # def test_argument_not_tags(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_tag=('mytag', 'othertag'))),
    #         '@e[tag=!mytag,tag=!othertag]')
    #
    # def test_argument_r(self):
    #     self.assertEqual(
    #         str(selectors.entities(r=90)),
    #         '@e[r=90.0]')
    #
    # def test_argument_rm(self):
    #     self.assertEqual(
    #         str(selectors.entities(rm=45)),
    #         '@e[rm=45.0]')
    #
    # def test_argument_rx(self):
    #     self.assertEqual(
    #         str(selectors.entities(rx=31.4159)),
    #         '@e[rx=31.4]')
    #
    # def test_argument_rxm(self):
    #     self.assertEqual(
    #         str(selectors.entities(rxm=27.1828)),
    #         '@e[rxm=27.2]')
    #
    # def test_argument_ry(self):
    #     self.assertEqual(
    #         str(selectors.entities(ry=31.4159)),
    #         '@e[ry=31.4]')
    #
    # def test_argument_rym(self):
    #     self.assertEqual(
    #         str(selectors.entities(rym=27.1828)),
    #         '@e[rym=27.2]')
    #
    # def test_argument_nbt(self):
    #     self.assertEqual(
    #         str(selectors.entities(nbt={'CustomName': 'Spacey Spacington'})),
    #         '@e[nbt={CustomName:"Spacey Spacington"}]')
    #
    # def test_argument_nbts(self):
    #     self.assertEqual(
    #         str(selectors.entities(nbt=({'CustomName': 'Spacey Spacington'}, {'NoAI': True}))),
    #         '@e[nbt={CustomName:"Spacey Spacington"},nbt={NoAI:1b}]')
    #
    # def test_argument_not_nbt(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_nbt={'CustomName': 'Spacey Spacington'})),
    #         '@e[nbt=!{CustomName:"Spacey Spacington"}]')
    #
    # def test_argument_not_nbts(self):
    #     self.assertEqual(
    #         str(selectors.entities(not_nbt=({'CustomName': 'Spacey Spacington'}, {'NoAI': True}))),
    #         '@e[nbt=!{CustomName:"Spacey Spacington"},nbt=!{NoAI:1b}]')
    #
    # def test_argument_c(self):
    #     self.assertEqual(
    #         str(selectors.entities(c=1)),
    #         '@e[c=1]')
    #
    # def test_everything(self):
    #     self.assertEqual(
    #         str(selectors.entities(
    #             position=Position(1.5, 2, 3.5),
    #             volume=Position(4, 9.5, 16),
    #             type='area_effect_cloud',
    #             l=10,
    #             lm=5,
    #             m='survival',
    #             team='myteam',
    #             max_scores={'myobjective': 100},
    #             min_scores={'myobjective': 1},
    #             name='myname',
    #             tag='mytag',
    #             r=90,
    #             rm=45,
    #             rx=31.4159,
    #             rxm=27.1828,
    #             ry=31.4159,
    #             rym=27.1828,
    #             nbt={'CustomName': 'Spacey Spacington'},
    #             c=1
    #         )),
    #         '@e[x=1.5,y=2.0,z=3.5,dx=4.0,dy=9.5,dz=16.0,type=area_effect_cloud,l=10,lm=5,m=survival,team=myteam,'
    #         'score_myobjective=100,score_myobjective_min=1,name=myname,tag=mytag,'
    #         'r=90.0,rm=45.0,rx=31.4,rxm=27.2,ry=31.4,rym=27.2,nbt={CustomName:"Spacey Spacington"},c=1]')



from .player_selector import player, PLAYER
from .random_selector import random, RANDOM
from .all_players_selector import all_players, ALL_PLAYERS
from .entities_selector import entities, ENTITIES
from .self_selector import self, SELF



import argparse
import logging
import unittest

import pymcutil.logging

parser = argparse.ArgumentParser(description='Run pymcutil tests.')
parser.add_argument('--verbosity', default=0, type=int, help='test runner verbosity')
parser.add_argument('--log', default='WARNING', help='log level')

args = parser.parse_args()

verbosity = args.verbosity
log_level = args.log

pymcutil.logging.install(log_level)

log = logging.getLogger(__name__)

unittest.TextTestRunner(verbosity=verbosity).run(unittest.TestLoader().discover('.'))



import unittest

from pymcutil.data_tag.compound_data_tag import CompoundDataTag
from pymcutil.data_tag.string_data_tag import StringDataTag


class CompoundDataTagTestCase(unittest.TestCase):
    def test_get(self):
        dt = CompoundDataTag(mykey='myvalue')
        self.assertEqual(dt['mykey'], StringDataTag('myvalue'))

    def test_set(self):
        dt = CompoundDataTag()
        dt['mykey'] = 'myvalue'
        self.assertEqual(dt['mykey'], StringDataTag('myvalue'))

    def test_overwrite(self):
        dt = CompoundDataTag(mykey='myvalue')
        dt['mykey'] = 'anothervalue'
        self.assertEqual(dt['mykey'], StringDataTag('anothervalue'))

    def test_del_only(self):
        dt = CompoundDataTag(mykey='myvalue')
        self.assertEqual(dt.to_str(), '{mykey:myvalue}')
        del dt['mykey']
        self.assertEqual(dt.to_str(), '{}')

    def test_del_with_remaining(self):
        dt = CompoundDataTag(mykey='myvalue', anotherkey='anothervalue')
        self.assertEqual(dt.to_str(), '{mykey:myvalue,anotherkey:anothervalue}')
        del dt['mykey']
        self.assertEqual(dt.to_str(), '{anotherkey:anothervalue}')

    def test_len(self):
        dt = CompoundDataTag(mykey='myvalue', anotherkey='anothervalue')
        self.assertEqual(len(dt), 2)

    def test_default(self):
        dt = CompoundDataTag()
        self.assertEqual(dt.to_str(), '{}')

    def test_empty_dict(self):
        dt = CompoundDataTag({})
        self.assertEqual(dt.to_str(), '{}')

    def test_multi_empty_dicts(self):
        dt = CompoundDataTag({}, {})
        self.assertEqual(dt.to_str(), '{}')

    def test_nonempty_dict_and_kwargs(self):
        dt = CompoundDataTag({'mykey': 'myvalue'}, anotherkey='anothervalue')
        self.assertEqual(dt.to_str(), '{mykey:myvalue,anotherkey:anothervalue}')

    def test_nonempty_dict_and_kwargs_overwriting(self):
        dt = CompoundDataTag({'mykey': 'myvalue'}, mykey='anothervalue')
        self.assertEqual(dt.to_str(), '{mykey:anothervalue}')

    def test_key_and_value(self):
        dt = CompoundDataTag(mykey='myvalue')
        self.assertEqual(dt.to_str(), '{mykey:myvalue}')

    def test_breaking_key_and_value(self):
        dt = CompoundDataTag({'my:key': 'somevalue'})
        self.assertEqual(dt.to_str(), '{"my:key":somevalue}')

    def test_key_and_breaking_value(self):
        dt = CompoundDataTag({'mykey': 'some:value'})
        self.assertEqual(dt.to_str(), '{mykey:"some:value"}')

    def test_breaking_key_and_breaking_value(self):
        dt = CompoundDataTag({'my:key': 'some:value'})
        self.assertEqual(dt.to_str(), '{"my:key":"some:value"}')

    def test_empty_value(self):
        dt = CompoundDataTag(mykey='')
        self.assertEqual(dt.to_str(), '{mykey:""}')

    def test_questionable_keys(self):
        dt = CompoundDataTag({'ke_y': 'value1', 'k.e--+_y': 'value2'})
        self.assertEqual(dt.to_str(), '{ke_y:value1,k.e--+_y:value2}')

    def test_questionable_values(self):
        dt = CompoundDataTag(key1='v__alue', key2='v.-_+alue')
        self.assertEqual(dt.to_str(), '{key1:v__alue,key2:v.-_+alue}')

    def test_breaking_keys(self):
        dt = CompoundDataTag({'ke$y': 'value1', 'ke y': 'value2', 'ke:y': 'value3'})
        self.assertEqual(dt.to_str(), '{"ke$y":value1,"ke y":value2,"ke:y":value3}')

    def test_breaking_values(self):
        dt = CompoundDataTag(key1='va lue', key2='va$lue', key3='minecraft:stone')
        self.assertEqual(dt.to_str(), '{key1:"va lue",key2:"va$lue",key3:"minecraft:stone"}')

    def test_key_with_quotes(self):
        dt = CompoundDataTag({'my"key': 'myvalue'})
        self.assertEqual(dt.to_str(), r'{"my\"key":myvalue}')

    def test_value_with_quotes(self):
        dt = CompoundDataTag(mykey='my"value')
        self.assertEqual(dt.to_str(), r'{mykey:"my\"value"}')

    def test_byte_data_tag_conversion(self):
        dt = CompoundDataTag(mybyte=True)
        self.assertEqual(dt.to_str(), '{mybyte:1b}')

    def test_int_data_tag_conversion(self):
        dt = CompoundDataTag(myint=123456789)
        self.assertEqual(dt.to_str(), '{myint:123456789}')

    def test_double_data_tag_conversion(self):
        dt = CompoundDataTag(mydouble=123.5)
        self.assertEqual(dt.to_str(), '{mydouble:123.5d}')

    def test_string_data_tag_conversion(self):
        dt = CompoundDataTag(mystring='hello')
        self.assertEqual(dt.to_str(), '{mystring:hello}')

    def test_list_data_tag_conversion(self):
        dt = CompoundDataTag(mylist=[])
        self.assertEqual(dt.to_str(), '{mylist:[]}')

    def test_compound_data_tag_conversion(self):
        dt = CompoundDataTag(mycompound={})
        self.assertEqual(dt.to_str(), '{mycompound:{}}')

    def test_deep_compound_data_tag_conversion(self):
        dt = CompoundDataTag(one=dict(two={'three': dict(four='five')}))
        self.assertEqual(dt.to_str(), '{one:{two:{three:{four:five}}}}')

    def test_all_conversions(self):
        dt = CompoundDataTag(
            mybyte=True,
            myint=123456789,
            mydouble=123.5,
            mystring='somestring',
            myemptylist=[],
            myemptycompound={},
            mycompound=dict(
                myinnerbyte=False,
                myinnerint=987654321,
                myinnerdouble=321.5,
                myinnerstring='someinnerstring',
                myinneremptycompound=dict()))
        self.assertEqual(
            dt.to_str(),
            '{mybyte:1b,myint:123456789,mydouble:123.5d,mystring:somestring,myemptylist:[],myemptycompound:{},'
            'mycompound:{myinnerbyte:0b,myinnerint:987654321,myinnerdouble:321.5d,myinnerstring:someinnerstring,'
            'myinneremptycompound:{}}}')



from pymcutil.constants import BYTE_MIN, BYTE_MAX
from pymcutil.data_tag import DataTag


class ByteDataTag(DataTag):
    """ A byte data tag, also known as TAG_Byte. """

    def __init__(self, value=0):
        self._value: int = None
        self.value = value

    def __eq__(self, other):
        return isinstance(other, ByteDataTag) and self.value.__eq__(other.value)

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, value):
        value = int(value)
        if not (BYTE_MIN <= value <= BYTE_MAX):
            raise ValueError(f'Value exceeds capacity of a byte: {value}')
        self._value = value

    def to_str(self) -> str:
        return f'{self.value}b'



from typing import Union

from pymcutil.block.block import Block
from pymcutil.position.position import Position
from .._execute_command import ExecuteCommand


class ExecuteConditionalBlockMixin(object):
    def __init__(self, position: Position.Generic, block: Block.Generic, subcommand: ExecuteCommand = None):
        self.position: Position = Position.sift(position)
        self.block: Block = Block.sift(block)
        self.subcommand: Union[ExecuteCommand, None] = subcommand



from pymcutil.command.command import Command
from .._execute_command import ExecuteCommand

CMD = 'then'


class ExecuteThenCommand(ExecuteCommand):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute then ...

    http://minecraft.gamepedia.com/Commands#execute
    """

    def __init__(self, command: Command):
        self.command: Command = command

    def params(self):
        yield from super().params()
        yield from (CMD, self.command)


then = ExecuteThenCommand



DATA_DIRECTORY = 'data'

ADVANCEMENTS_DIRECTORY = 'advancements'
FUNCTIONS_DIRECTORY = 'functions'
LOOT_TABLES_DIRECTORY = 'loot_tables'

ADVANCEMENTS_EXTENSION = 'json'
FUNCTIONS_EXTENSION = 'mcfunction'
LOOT_TABLES_EXTENSION = 'json'



from typing import Iterable

from pymcutil.command import commands
from pymcutil.command.command import Command
from pymcutil.resource.resource_generator.abc.procedure import Procedure
from pymcutil.resource.resource_manager.abc.resource_manager import ResourceManager
from . import functions


def split_in_two(s):
    index = len(s) // 2
    s1, s2 = s[:index], s[index:]
    return s1, s2


class StringSplitProcedure(Procedure):
    def commands(
            self, manager: ResourceManager, reference: functions.StringSplitFunctionResourceReference) -> Iterable[Command]:
        message, indent = reference.params
        yield commands.say(message='. . ' * indent + message)
        if len(message) > 1:
            former, latter = split_in_two(message)
            yield commands.function(function=manager.name(functions.string_split(former, indent + 1)))
            yield commands.function(function=manager.name(functions.string_split(latter, indent + 1)))


string_split = StringSplitProcedure



from typing import Iterable, List, Union

from pymcutil.util.siftable import SequenceSiftable


class Vector(Iterable, SequenceSiftable):
    Generic = Union['Vector', Iterable]

    def __init__(self, *components):
        # Don't cast potential subtypes of float into normal floats!
        self._components = [c if isinstance(c, float) else float(c) for c in components]

    def __repr__(self):
        return '{name}({params})'.format(
            name=self.__class__.__name__,
            params=', '.join(['{!r}'.format(c) for c in self.components]))

    def __str__(self):
        return self.to_str()

    def __eq__(self, other):
        if type(other) is type(self) and len(self.components) == len(other.components):
            for c1, c2 in zip(self.components, other.components):
                if c1 != c2:
                    return False
            return True
        return False

    def __add__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return self.__class__(*[c + other for c in self.components])
        elif isinstance(other, type(self)) and len(self.components) == len(other.components):
            return self.__class__(*[c1 + c2 for c1, c2 in zip(self.components, other.components)])
        raise TypeError('{self!r} cannot add with {other!r}'.format(self=self, other=other))

    def __sub__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return self.__class__(*[c - other for c in self.components])
        elif isinstance(other, type(self)) and len(self.components) == len(other.components):
            return self.__class__(*[c1 - c2 for c1, c2 in zip(self.components, other.components)])
        raise TypeError('{self!r} cannot sub with {other!r}'.format(self=self, other=other))

    def __mul__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return self.__class__(*[c * other for c in self.components])
        elif isinstance(other, type(self)) and len(self.components) == len(other.components):
            return self.__class__(*[c1 * c2 for c1, c2 in zip(self.components, other.components)])
        raise TypeError('{self!r} cannot mul with {other!r}'.format(self=self, other=other))

    def __truediv__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return self.__class__(*[c / other for c in self.components])
        elif isinstance(other, type(self)) and len(self.components) == len(other.components):
            return self.__class__(*[c1 / c2 for c1, c2 in zip(self.components, other.components)])
        raise TypeError('{self!r} cannot truediv with {other!r}'.format(self=self, other=other))

    def __floordiv__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return self.__class__(*[c // other for c in self.components])
        elif isinstance(other, type(self)) and len(self.components) == len(other.components):
            return self.__class__(*[c1 // c2 for c1, c2 in zip(self.components, other.components)])
        raise TypeError('{self!r} cannot floordiv with {other!r}'.format(self=self, other=other))

    def __iadd__(self, other):
        return self.__add__(other)

    def __isub__(self, other):
        return self.__sub__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __itruediv__(self, other):
        return self.__truediv__(other)

    def __ifloordiv__(self, other):
        return self.__floordiv__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return self.__mul__(-1)

    def __iter__(self):
        return self.components.__iter__()

    def to_str(self) -> str:
        return ' '.join(['{:g}'.format(c) for c in self.components])

    def to_int_str(self) -> str:
        return ' '.join(['{:d}'.format(c) for c in self.to_int_list()])

    def to_float_str(self) -> str:
        return ' '.join(['{:g}'.format(c) for c in self.to_float_list()])

    def to_int_list(self) -> List[int]:
        return [int(c) for c in self.components]

    def to_float_list(self) -> List[float]:
        return [float(c) for c in self.components]

    def expand(self, other):
        if isinstance(other, type(self)) and len(self.components) == len(other.components):
            return self.__class__(*[max(c1, c2) for c1, c2 in zip(self.components, other.components)])
        raise TypeError('{self!r} cannot expand with {other!r}'.format(self=self, other=other))

    def contract(self, other):
        if isinstance(other, type(self)) and len(self.components) == len(other.components):
            return self.__class__(*[min(c1, c2) for c1, c2 in zip(self.components, other.components)])
        raise TypeError('{self!r} cannot contract with {other!r}'.format(self=self, other=other))

    @property
    def components(self) -> List[float]:
        return self._components

    @property
    def magnitude(self) -> float:
        m = 1.0
        for c in self.components:
            m *= c
        return m



from pymcutil.command.commands.effect_command import EffectCommand
from pymcutil.command.target import Target

CMD = 'give'


class EffectGiveCommand(EffectCommand):
    """
    An objective model of one of Minecraft's `effect` subcommands:

        effect give <entity> <effect> [seconds] [amplifier] [hideParticles]

    http://minecraft.gamepedia.com/Commands#effect
    """

    def __init__(
            self, target: Target, effect: str, seconds: int = None, amplifier: int = None, hide_particles: bool = None):
        self.target: Target = target
        self.effect: str = effect
        self.seconds: int = seconds
        self.amplifier: int = amplifier
        self.hide_particles = hide_particles

    def params(self):
        yield from super().params()
        yield from (
            CMD, self.target, self.effect,
            (self.seconds, 1000000),
            (self.amplifier, 0),
            (self.hide_particles, True))


give = EffectGiveCommand



import unittest

from pymcutil.data_tag.byte_data_tag import ByteDataTag
from pymcutil.data_tag.compound_data_tag import CompoundDataTag
from pymcutil.data_tag.double_data_tag import DoubleDataTag
from pymcutil.data_tag.float_data_tag import FloatDataTag
from pymcutil.data_tag.int_data_tag import IntDataTag
from pymcutil.data_tag.list_data_tag import ListDataTag
from pymcutil.data_tag.long_data_tag import LongDataTag
from pymcutil.data_tag.short_data_tag import ShortDataTag
from pymcutil.data_tag.string_data_tag import StringDataTag


class CompoundDataTagTestCase(unittest.TestCase):
    def test_everything(self):
        dt = CompoundDataTag(
            explicitbyte=ByteDataTag(True),
            explicitshort=ShortDataTag(12),
            explicitint=IntDataTag(123),
            explicitlong=LongDataTag(1234),
            explicitfloat=FloatDataTag(1234.5),
            explicitdouble=DoubleDataTag(1234.56),
            explicitstring=StringDataTag('hello'),
            explicitlist=ListDataTag([False, 9001, 'item']),
            explicitcompound=CompoundDataTag({'key1': True, 'key2': 25565, 'key3': 'stringvalue'}),
            anotherexplicitcompound=CompoundDataTag(key4=False, key5=25566, key6='stringvalue2'),
            implicitbyte=False,
            implicitint=321,
            implicitdouble=6543.21,
            implicitstring='goodbye',
            implicitlist=[True, 9002, 'anotheritem'],
            implicitcompound={'key7': True, 'key8': 25567, 'key9': 'stringvalue3'},
            anotherimplicitcompound=dict(key10=False, key11=25568, key12='stringvalue4'),
            breakingstring='breaking:string',
            quotedstring='quoted"string',
            innerbreakingstring=['inner:breakingstring'],
            innerquotedstring=['inner"quotedstring'],
            deepbreakingstring=[dict(one=[dict(two=[dict(three='deep:breakingstring')])])],
            deepquotedstring=[dict(one=[dict(two=[dict(three='deep"quotedstring')])])])
        self.assertEqual(
            dt.to_str(),
            r"""
{explicitbyte:1b,explicitshort:12s,explicitint:123,explicitlong:1234l,explicitfloat:1234.5f,explicitdouble:1234.56d,
explicitstring:hello,explicitlist:[0b,9001,item],explicitcompound:{key1:1b,key2:25565,key3:stringvalue},
anotherexplicitcompound:{key4:0b,key5:25566,key6:stringvalue2},implicitbyte:0b,implicitint:321,implicitdouble:6543.21d,
implicitstring:goodbye,implicitlist:[1b,9002,anotheritem],implicitcompound:{key7:1b,key8:25567,key9:stringvalue3},
anotherimplicitcompound:{key10:0b,key11:25568,key12:stringvalue4},breakingstring:"breaking:string",
quotedstring:"quoted\"string",innerbreakingstring:["inner:breakingstring"],innerquotedstring:["inner\"quotedstring"],
deepbreakingstring:[{one:[{two:[{three:"deep:breakingstring"}]}]}],
deepquotedstring:[{one:[{two:[{three:"deep\"quotedstring"}]}]}]}
""".replace('\n', ''))



from typing import Union

from pymcutil.position.position import Position
from .._execute_command import ExecuteCommand

CMD = 'offset'


class ExecuteOffsetCommand(ExecuteCommand):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute offset <pos: x y z> -> execute

    http://minecraft.gamepedia.com/Commands#execute
    """

    def __init__(self, position: Position.Generic, subcommand: ExecuteCommand = None):
        self.position: Position = Position.sift(position)
        self.subcommand: Union[ExecuteCommand, None] = subcommand

    def params(self):
        yield from super().params()
        yield from (
            CMD, self.position,
            (self.subcommand.substr(), None))


offset = ExecuteOffsetCommand



import unittest

from pymcutil.position.vector_3d import Vector3D


class Vector3DTestCase(unittest.TestCase):
    def test_init(self):
        v = Vector3D(1, 2, 3)
        self.assertEqual(v.x, 1)
        self.assertEqual(v.y, 2)
        self.assertEqual(v.z, 3)

    def test_str(self):
        self.assertEqual(str(Vector3D(1, 2, 3)), '1 2 3')



from typing import Union

from pymcutil.command.target import Target
from .._execute_command import ExecuteCommand

CMD = 'as'


class ExecuteAsCommand(ExecuteCommand):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute as <targets: entities> -> execute

    http://minecraft.gamepedia.com/Commands#execute
    """

    def __init__(self, target: Target, subcommand: ExecuteCommand = None):
        self.target: Target = target
        self.subcommand: Union[ExecuteCommand, None] = subcommand

    def params(self):
        yield from super().params()
        yield from (
            CMD, self.target,
            (self.subcommand.substr(), None))


as_ = ExecuteAsCommand  # work around keyword conflict



import unittest

from pymcutil.position.vector import Vector


class VectorTestCase(unittest.TestCase):
    def test_repr(self):
        self.assertEqual(repr(Vector(1, 2)), 'Vector(1.0, 2.0)')

    def test_str(self):
        self.assertEqual(str(Vector(1, 2)), '1 2')

    def test_to_int_str(self):
        self.assertEqual(Vector(1, 2.5).to_int_str(), '1 2')

    def test_to_float_str(self):
        self.assertEqual(Vector(1, 2.5).to_float_str(), '1 2.5')

    def test_to_int_list(self):
        self.assertEqual(Vector(1, 2.5).to_int_list(), [1, 2])

    def test_to_float_list(self):
        self.assertEqual(Vector(1, 2.5).to_float_list(), [1.0, 2.5])

    # ---- eq ----------------------------------------------------------------------------------------------------------

    def test_eq(self):
        self.assertEqual(Vector(1, 2), Vector(1.0, 2.0))

    def test_not_eq(self):
        self.assertNotEqual(Vector(1, 2), Vector(1.0, 1.0))

    def test_really_not_eq(self):
        self.assertNotEqual(Vector(1, 2), Vector(2.0, 1.0))

    def test_definitely_not_eq(self):
        self.assertNotEqual(Vector(1, 2), Vector(1, 2, 3))

    def test_really_definitely_not_eq(self):
        self.assertNotEqual(Vector(1, 2), 'incompatible')

    # ---- add ---------------------------------------------------------------------------------------------------------

    def test_add_with_int(self):
        self.assertEqual(Vector(1, 2) + 1, Vector(2, 3))

    def test_add_with_float(self):
        self.assertEqual(Vector(1, 2) + 0.5, Vector(1.5, 2.5))

    def test_add_with_vector(self):
        self.assertEqual(Vector(1, 2) + Vector(2, 3), Vector(3, 5))

    def test_add_incompatible_type_error(self):
        with self.assertRaises(TypeError):
            Vector(1, 2) + 'incompatible'

    def test_add_incompatible_vector_error(self):
        with self.assertRaises(TypeError):
            Vector(1, 2) + Vector(1, 2, 3)

    # ---- sub ---------------------------------------------------------------------------------------------------------

    def test_sub_with_int(self):
        self.assertEqual(Vector(1, 2) - 1, Vector(0, 1))

    def test_sub_with_float(self):
        self.assertEqual(Vector(1, 2) - 0.5, Vector(0.5, 1.5))

    def test_sub_with_vector(self):
        self.assertEqual(Vector(3, 5) - Vector(2, 3), Vector(1, 2))

    def test_sub_incompatible_type_error(self):
        with self.assertRaises(TypeError):
            Vector(1, 2) - 'incompatible'

    def test_sub_incompatible_vector_error(self):
        with self.assertRaises(TypeError):
            Vector(3, 5) - Vector(2, 3, 4)

    # ---- mul ---------------------------------------------------------------------------------------------------------

    def test_mul_with_int(self):
        self.assertEqual(Vector(1, 2) * 2, Vector(2, 4))

    def test_mul_with_float(self):
        self.assertEqual(Vector(1, 2) * 0.5, Vector(0.5, 1))

    def test_mul_with_vector(self):
        self.assertEqual(Vector(1, 2) * Vector(2, 3), Vector(2, 6))

    def test_mul_incompatible_type_error(self):
        with self.assertRaises(TypeError):
            Vector(1, 2) * 'incompatible'

    def test_mul_incompatible_vector_error(self):
        with self.assertRaises(TypeError):
            Vector(1, 2) * Vector(2, 3, 4)

    # ---- truediv -----------------------------------------------------------------------------------------------------

    def test_truediv_with_int(self):
        self.assertEqual(Vector(7, 4) / 2, Vector(3.5, 2))

    def test_truediv_with_float(self):
        self.assertEqual(Vector(1.25, 2.1) / 0.5, Vector(2.5, 4.2))

    def test_truediv_with_vector(self):
        self.assertEqual(Vector(7, 4) / Vector(2, 5), Vector(3.5, 0.8))

    def test_truediv_incompatible_type_error(self):
        with self.assertRaises(TypeError):
            Vector(7, 4) / 'incompatible'

    def test_truediv_incompatible_vector_error(self):
        with self.assertRaises(TypeError):
            Vector(7, 4) / Vector(1, 2, 3)

    # ---- floordiv ----------------------------------------------------------------------------------------------------

    def test_floordive_with_int(self):
        self.assertEqual(Vector(7, 4) // 2, Vector(3, 2))

    def test_floordive_with_float(self):
        self.assertEqual(Vector(1.25, 2.1) // 0.5, Vector(2, 4))

    def test_floordive_with_vector(self):
        self.assertEqual(Vector(7, 4) // Vector(2, 5), Vector(3, 0))

    def test_floordive_incompatible_type_error(self):
        with self.assertRaises(TypeError):
            Vector(7, 4) // 'incompatible'

    def test_floordive_incompatible_vector_error(self):
        with self.assertRaises(TypeError):
            Vector(7, 4) // Vector(1, 2, 3)

    # ---- others ------------------------------------------------------------------------------------------------------

    def test_iadd(self):
        v = Vector(1, 2)
        v += Vector(2, 3)
        self.assertEqual(v, Vector(3, 5))

    def test_isub(self):
        rc = Vector(3, 5)
        rc -= Vector(2, 3)
        self.assertEqual(rc, Vector(1, 2))

    def test_imul(self):
        rc = Vector(1, 2)
        rc *= Vector(2, 3)
        self.assertEqual(rc, Vector(2, 6))

    def test_itruediv(self):
        rc = Vector(7, 4)
        rc /= Vector(2, 5)
        self.assertEqual(rc, Vector(3.5, 0.8))

    def test_ifloordiv(self):
        rc = Vector(7, 4)
        rc //= Vector(2, 5)
        self.assertEqual(rc, Vector(3, 0))

    def test_radd(self):
        with self.assertRaises(TypeError):
            'incompatible' + Vector(1, 2)

    def test_rmul(self):
        with self.assertRaises(TypeError):
            'incompatible' * Vector(1, 2)

    def test_neg(self):
        self.assertEqual(-Vector(1, 2), Vector(-1, -2))

    def test_iter(self):
        self.assertEqual([c for c in Vector(1, 2, 3)], [1, 2, 3])

    # ------------------------------------------------------------------------------------------------------------------

    def test_components(self):
        self.assertEqual(Vector(1, 2).components, [1, 2])

    def test_magnitude(self):
        self.assertEqual(Vector(2, 3, 0.5).magnitude, 3)

    def test_expand(self):
        self.assertEqual(Vector(1, 2).expand(Vector(3, 1)), Vector(3, 2))

    def test_expand_incompatible_type_error(self):
        with self.assertRaises(TypeError):
            Vector(1, 2).expand('incompatible')

    def test_expand_incompatible_vector_error(self):
        with self.assertRaises(TypeError):
            Vector(1, 2).expand(Vector(3, 1, 0))

    def test_contract(self):
        self.assertEqual(Vector(3, 2).contract(Vector(1, 3)), Vector(1, 2))

    def test_contract_incompatible_type_error(self):
        with self.assertRaises(TypeError):
            Vector(3, 2).contract('incompatible')

    def test_contract_incompatible_vector_error(self):
        with self.assertRaises(TypeError):
            Vector(3, 2).contract(Vector(1, 2, 0))



import unittest

from pymcutil.command import commands
from pymcutil.selector import selectors


class FunctionCommandTestCase(unittest.TestCase):
    def test_without_any_optionals(self):
        cmd = commands.function(
            function='foo')
        self.assertEqual(str(cmd), 'function foo')

    def test_default_optionals(self):
        cmd = commands.function(
            function='foo',
            mode_target=selectors.SELF)
        self.assertEqual(str(cmd), 'function foo if @s')

    def test_up_to_mode(self):
        cmd = commands.function(
            function='foo',
            mode='if')
        self.assertEqual(str(cmd), 'function foo if')

    def test_up_to_mode_target(self):
        cmd = commands.function(
            function='foo',
            mode='if',
            mode_target=selectors.SELF)
        self.assertEqual(str(cmd), 'function foo if @s')



from pymcutil.util.symbol import Symbol


class Gamemode(Symbol):
    pass


survival = Gamemode('survival')
adventure = Gamemode('adventure')
creative = Gamemode('creative')
spectator = Gamemode('spectator')



import unittest

from pymcutil.position.rc import rel
from pymcutil.position.rotation import Rotation, ZERO_ROTATION, ZERO_TURN


class RotationTestCase(unittest.TestCase):
    def test_init(self):
        r = Rotation(45, 90)
        self.assertEqual(r.yaw, 45)
        self.assertEqual(r.pitch, 90)

    def test_invert(self):
        self.assertEqual(str(~Rotation(45, 90)), '~45 ~90')

    def test_abs(self):
        self.assertEqual(str(abs(~Rotation(45, 90))), '45 90')

    def test_zero_rotation(self):
        self.assertEqual(str(ZERO_ROTATION), '0 0')

    def test_zero_turn(self):
        self.assertEqual(str(ZERO_TURN), '~ ~')

    def test_rel_mixed_zero_coords(self):
        self.assertEqual(str(~Rotation(0, 30)), '~ ~30')

    def test_rel_decimal_coords(self):
        self.assertEqual(str(~Rotation(3.142, 6.674)), '~3.142 ~6.674')

    def test_mixed_abs_and_rel_coords(self):
        self.assertEqual(str(Rotation(45, rel(30))), '45 ~30')

    def test_mixed_abs_and_rel_to_rel_coords(self):
        self.assertEqual(str(~Rotation(45, rel(90))), '~45 ~90')



import unittest

from pymcutil.position.vector_2d import Vector2D


class Vector2DTestCase(unittest.TestCase):
    def test_init(self):
        v = Vector2D(1, 2)
        self.assertEqual(v.x, 1)
        self.assertEqual(v.y, 2)

    def test_str(self):
        self.assertEqual(str(Vector2D(1, 2)), '1 2')



from .._execute_command import ExecuteCommand

CMD = 'unless'


class ExecuteUnlessCommand(ExecuteCommand):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute unless ...

    http://minecraft.gamepedia.com/Commands#execute

    *This command requires a sub-command to be useful.*
    """

    def params(self):
        yield CMD



import abc


class Resource(abc.ABC):
    """ Generic game resource representation, used to define common access methods. """

    @abc.abstractmethod
    def write_to(self, path):
        """ Write this `Resource` to file at `path`. """



import unittest

from pymcutil.position.rc import RC, rel


class RCTestCase(unittest.TestCase):
    def test_rel(self):
        self.assertEqual(rel(1), RC(1))

    def test_repr(self):
        self.assertEqual(repr(RC(1)), 'RC(1.0)')

    def test_str(self):
        self.assertEqual(str(RC(1)), '~1.0')

    def test_zero_str(self):
        self.assertEqual(str(RC(0)), '~')

    def test_eq(self):
        self.assertEqual(RC(1), RC(1.0))

    # TODO Decide whether a relative coord should be equal to an absolute coord.
    # Probably not, but it seems we have an issue with inheriting float in that comparing to any builtin value
    # will fallback to float's equivalency method, overriding ours. Perhaps we need to componentize the float instead.

    # def test_not_eq_to_float(self):
    #     self.assertNotEqual(RC(1), 1.0)
    #
    # def test_not_eq_to_int(self):
    #     self.assertNotEqual(RC(1), 1)

    def test_not_eq_to_str(self):
        self.assertNotEqual(RC(1), 'incompatible')

    def test_add(self):
        self.assertEqual(RC(1) + RC(2), RC(3))

    def test_add_error(self):
        with self.assertRaises(TypeError):
            RC(1) + 'incompatible'

    def test_sub(self):
        self.assertEqual(RC(3) - RC(2), RC(1))

    def test_sub_error(self):
        with self.assertRaises(TypeError):
            RC(3) - 'incompatible'

    def test_mul(self):
        self.assertEqual(RC(3) * RC(2), RC(6))

    def test_mul_error(self):
        with self.assertRaises(TypeError):
            RC(3) * 'incompatible'

    def test_truediv(self):
        self.assertEqual(RC(7) / RC(2), RC(3.5))

    def test_truediv_error(self):
        with self.assertRaises(TypeError):
            RC(7) / 'incompatible'

    def test_floordiv(self):
        self.assertEqual(RC(7) // RC(2), RC(3))

    def test_floordiv_error(self):
        with self.assertRaises(TypeError):
            RC(7) // 'incompatible'

    def test_iadd(self):
        rc = RC(1)
        rc += RC(2)
        self.assertEqual(rc, RC(3))

    def test_isub(self):
        rc = RC(3)
        rc -= RC(2)
        self.assertEqual(rc, RC(1))

    def test_imul(self):
        rc = RC(3)
        rc *= RC(2)
        self.assertEqual(rc, RC(6))

    def test_itruediv(self):
        rc = RC(7)
        rc /= RC(2)
        self.assertEqual(rc, RC(3.5))

    def test_ifloordiv(self):
        rc = RC(7)
        rc //= RC(2)
        self.assertEqual(rc, RC(3))

    def test_radd(self):
        with self.assertRaises(TypeError):
            'incompatible' + RC(2)

    def test_rmul(self):
        with self.assertRaises(TypeError):
            'incompatible' * RC(2)

    def test_neg(self):
        self.assertEqual(-RC(1), RC(-1))

    # The resulting type of an operation between an absolute and relative coordinate should match whichever came first.

    def test_rel_add_abs_str(self):
        self.assertEqual(str(RC(1) + 2.0), '~3.0')

    def test_abs_add_rel_str(self):
        self.assertEqual(str(1.0 + RC(2)), '3.0')

    def test_rel_sub_abs_str(self):
        self.assertEqual(str(RC(3) - 2.0), '~1.0')

    def test_abs_sub_rel_str(self):
        self.assertEqual(str(3.0 - RC(2)), '1.0')

    def test_rel_mul_abs_str(self):
        self.assertEqual(str(RC(2) * 3.0), '~6.0')

    def test_abs_mul_rel_str(self):
        self.assertEqual(str(2.0 * RC(3)), '6.0')

    def test_rel_truediv_abs_str(self):
        self.assertEqual(str(RC(7) / 2), '~3.5')

    def test_abs_truediv_rel_str(self):
        self.assertEqual(str(7.0 / RC(2)), '3.5')

    def test_rel_floordiv_abs_str(self):
        self.assertEqual(str(RC(7) // 2), '~3.0')

    def test_abs_floordiv_rel_str(self):
        self.assertEqual(str(7.0 // RC(2)), '3.0')



from pymcutil.position.rc import RC
from pymcutil.position.vector_3d import Vector3D


class Position(Vector3D):
    """ Position AKA location representation containing x, y, z coordinates that may be relative. Note that all
    position operations are defined under the inherited Vector class, meaning we can conveniently add, subtract, etc
    positions from one another. We can also make a position relative by invoking the ~ operator on it, or turn
    individual coordinates relative by explicitly using the RC class. """

    @classmethod
    def _from_other(cls, obj):
        # A tuple `(x, y, z)` will produce relative coordinates, whereas everything else (such as a list `[x, y, z]`)
        # will instead produce absolute coordinates.
        sifted = super()._from_other(obj)
        return ~sifted if isinstance(obj, tuple) else sifted

    def __init__(self, x: float, y: float, z: float):
        super().__init__(x, y, z)

    def __invert__(self):
        return self.relative()

    def __abs__(self):
        return self.absolute()

    def relative(self):
        return self.__class__(*[RC(c) for c in self.components])

    def absolute(self):
        return self.__class__(*[float(c) for c in self.components])


ZERO_POSITION = Position(0, 0, 0)
ZERO_OFFSET = ~ZERO_POSITION
BELOW_OFFSET = ~Position(0, -1, 0)
ABOVE_OFFSET = ~Position(0, 1, 0)



from pymcutil.command.commands.effect_command import EffectCommand
from pymcutil.command.target import Target

CMD = 'take'


class EffectTakeCommand(EffectCommand):
    """
    An objective model of one of Minecraft's `effect` subcommands:

        effect take <entity> [<effect>]

    http://minecraft.gamepedia.com/Commands#effect
    """

    def __init__(self, target: Target, effect: str = None):
        self.target: Target = target
        self.effect: str = effect

    def params(self):
        yield from super().params()
        yield from (
            CMD, self.target,
            (self.effect, '*'))


take = EffectTakeCommand



from pymcutil.constants import INT_MIN, INT_MAX
from pymcutil.data_tag import DataTag


class IntDataTag(DataTag):
    """ An int data tag, also known as TAG_Int. """

    def __init__(self, value=0):
        self._value: int = None
        self.value = value

    def __eq__(self, other):
        return isinstance(other, IntDataTag) and self.value.__eq__(other.value)

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, value):
        value = int(value)
        if not (INT_MIN <= value <= INT_MAX):
            raise ValueError(f'Value exceeds capacity of an int: {value}')
        self._value = value

    def to_str(self) -> str:
        # Ints are the default data tag for any integer value, and lack a suffix entirely.
        return f'{self.value}'



from typing import Callable, Optional


def nullable(obj: object, converter: Callable):
    return converter(obj) if obj is not None else None


def nullable_float(obj: object) -> Optional[float]:
    return nullable(obj, float)


def nullable_int(obj: object) -> Optional[int]:
    return nullable(obj, int)


def nullable_bool(obj: object) -> Optional[bool]:
    return nullable(obj, bool)


def nullable_str(obj: object) -> Optional[str]:
    return nullable(obj, str)


def nullable_tuple(obj: object, converter: Callable) -> Optional[tuple]:
    t = nullable(obj, tuple)
    return None if t is None else tuple(converter(item) for item in t)



from collections import Mapping, MutableMapping

from pymcutil.data_tag import DataTag
from pymcutil.data_tag.helper import data_taggify
from pymcutil.data_tag.string_data_tag import StringDataTag
from pymcutil.util.siftable import MappingSiftable


class CompoundDataTag(DataTag, MutableMapping, MappingSiftable):
    """ A compound data tag, also known as TAG_Compound. """

    Generic = Mapping

    def __init__(self, *args, **kwargs):
        self._value = {}

        # Assume all unnamed args are mappings that can be inherited from.
        for arg in args:
            self.update(**arg)

        # Update from all named args.
        self.update(kwargs)

    def __delitem__(self, key):
        return self._value.__delitem__(key)

    def __iter__(self):
        return self._value.__iter__()

    def __setitem__(self, key, value):
        key = str(key)
        value = data_taggify(value)
        return self._value.__setitem__(key, value)

    def __len__(self):
        return self._value.__len__()

    def __getitem__(self, key):
        return self._value.__getitem__(key)

    def __eq__(self, other):
        return isinstance(other, CompoundDataTag) and self.value.__eq__(other.value)

    @property
    def value(self) -> dict:
        return self._value

    def to_str(self) -> str:
        # Keys are strings and use the same syntax rules as NBT strings.
        pairs = (f'{StringDataTag(key)}:{value}' for key, value in self.items())
        return ''.join(('{', ','.join(pairs), '}'))



import abc
from numbers import Real
from typing import Iterable, Mapping, Optional, Tuple, Union

from pymcutil.data_tag.compound_data_tag import CompoundDataTag
from pymcutil.position.position import Position
from pymcutil.selector.advancement_set import AdvancementSet
from pymcutil.selector.range import Range
from pymcutil.selector.score_set import ScoreSet
from pymcutil.symbols.entity_ids.entity_ids import EntityID
from pymcutil.symbols.gamemodes.gamemode import Gamemode
from pymcutil.symbols.selector_bases.selector_bases import SelectorBase
from pymcutil.symbols.selector_sorts.selector_sort import SelectorSort
from pymcutil.util import first
from pymcutil.util.nullability import nullable_float, nullable_int, nullable_str, nullable_tuple
from pymcutil.util.siftable import MappingSiftable

RepeatableString = Iterable[str]
RepeatableEntityID = Iterable[EntityID.Generic]
RepeatableGamemode = Iterable[Gamemode.Generic]
RepeatableNBT = Iterable[CompoundDataTag.Generic]


class Selector(MappingSiftable):
    """
    Represents a Minecraft target selector, with similar arguments and some built-in conveniences.

    Provided arguments cascade such that the most specific value is used. For example, an explicitly provided `x`
    coordinate will take precedence over the `x` component of any given `position`.
    """

    Generic = Union['Selector', Mapping]

    def __init__(
            self,
            position: Position.Generic = None, x: Real = None, y: Real = None, z: Real = None,
            volume: Position.Generic = None, dx: Real = None, dy: Real = None, dz: Real = None,
            x_rotation: Range.Generic = None, y_rotation: Range.Generic = None,
            distance: Range.Generic = None,
            level: Range.Generic = None,
            type_: EntityID = None, not_types: RepeatableEntityID = None,
            name: str = None, not_names: RepeatableString = None,
            team: str = None, not_teams: RepeatableString = None,
            gamemode: Gamemode.Generic = None, not_gamemodes: RepeatableGamemode = None,
            tags: RepeatableString = None, not_tags: RepeatableString = None,
            nbts: RepeatableNBT = None, not_nbts: RepeatableNBT = None,
            scores: ScoreSet.Generic = None,
            advancements: AdvancementSet.Generic = None,
            sort: SelectorSort = None,
            limit: int = None,
    ):
        # Intermediates used for calculations.

        position = Position.sift(position, None)
        volume = Position.sift(volume, None)

        x = first(x, position.x if isinstance(position, Position) else None)
        y = first(y, position.y if isinstance(position, Position) else None)
        z = first(z, position.z if isinstance(position, Position) else None)

        dx = first(dx, volume.x if isinstance(volume, Position) else None)
        dy = first(dy, volume.y if isinstance(volume, Position) else None)
        dz = first(dz, volume.z if isinstance(volume, Position) else None)

        # Actual instance attributes.

        self.x: Optional[float] = nullable_float(x)
        self.y: Optional[float] = nullable_float(y)
        self.z: Optional[float] = nullable_float(z)

        self.dx: Optional[float] = nullable_float(dx)
        self.dy: Optional[float] = nullable_float(dy)
        self.dz: Optional[float] = nullable_float(dz)

        self.x_rotation: Optional[Range] = Range.sift(x_rotation, None)
        self.y_rotation: Optional[Range] = Range.sift(y_rotation, None)
        self.distance: Optional[Range] = Range.sift(distance, None)
        self.level: Optional[Range] = Range.sift(level, None)

        self.type_: Optional[EntityID] = nullable_str(type_)
        self.not_types: Optional[Tuple[EntityID]] = nullable_tuple(not_types, converter=EntityID.sift)

        self.name: Optional[str] = nullable_str(name)
        self.not_names: Optional[Tuple[str]] = nullable_tuple(not_names, converter=str)

        self.team: Optional[str] = nullable_str(team)
        self.not_teams: Optional[Tuple[str]] = nullable_tuple(not_teams, converter=str)

        self.gamemode: Optional[Gamemode] = Gamemode.sift(gamemode, None)
        self.not_gamemodes: Optional[Tuple[Gamemode]] = nullable_tuple(not_gamemodes, converter=Gamemode.sift)

        self.tags: Optional[Tuple[str]] = nullable_tuple(tags, converter=str)
        self.not_tags: Optional[Tuple[str]] = nullable_tuple(not_tags, converter=str)

        self.nbts: Optional[Tuple[CompoundDataTag]] = nullable_tuple(nbts, converter=CompoundDataTag.sift)
        self.not_nbts: Optional[Tuple[CompoundDataTag]] = nullable_tuple(not_nbts, converter=CompoundDataTag.sift)

        self.scores: Optional[ScoreSet] = ScoreSet.sift(scores, None)

        self.advancements: Optional[AdvancementSet] = AdvancementSet.sift(advancements, None)

        self.sort: Optional[SelectorSort] = SelectorSort.sift(sort, None)

        self.limit: Optional[int] = nullable_int(limit)

    def __str__(self):
        arguments = tuple(self.arguments)
        argument_str = ''.join(('[', ','.join(['{}={}'.format(k, v) for k, v in arguments]), ']')) if arguments else ''
        return '@{}{}'.format(self.base, argument_str or '')

    @property
    def arguments(self) -> Iterable[Tuple[str, str]]:
        # TODO This is disguesting. Has it always been this way?
        if self.x is not None:
            yield 'x', str(self.x)
        if self.y is not None:
            yield 'y', str(self.y)
        if self.z is not None:
            yield 'z', str(self.z)

        if self.dx is not None:
            yield 'dx', str(self.dx)
        if self.dy is not None:
            yield 'dy', str(self.dy)
        if self.dz is not None:
            yield 'dz', str(self.dz)

        if self.x_rotation is not None:
            yield 'x_rotation', str(self.x_rotation)
        if self.y_rotation is not None:
            yield 'y_rotation', str(self.y_rotation)

        if self.distance is not None:
            yield 'distance', str(self.distance)

        if self.level is not None:
            yield 'level', str(self.level)

        if self.type_ is not None:
            yield 'type', str(self.type_)
        if self.not_types is not None:
            yield from (('type', '!{}'.format(type_)) for type_ in self.not_types)

        if self.name is not None:
            yield 'name', str(self.name)
        if self.not_names is not None:
            yield from (('name', '!{}'.format(name)) for name in self.not_names)

        if self.team is not None:
            yield 'team', str(self.team)
        if self.not_teams is not None:
            yield from (('team', '!{}'.format(team)) for team in self.not_teams)

        if self.gamemode is not None:
            yield 'gamemode', str(self.gamemode)
        if self.not_gamemodes is not None:
            yield from (('gamemode', '!{}'.format(gamemode)) for gamemode in self.not_gamemodes)

        if self.tags is not None:
            yield from (('tag', '{}'.format(tag)) for tag in self.tags)
        if self.not_tags is not None:
            yield from (('tag', '!{}'.format(tag)) for tag in self.not_tags)

        if self.nbts is not None:
            yield from (('nbt', '{}'.format(nbt)) for nbt in self.nbts)
        if self.not_nbts is not None:
            yield from (('nbt', '!{}'.format(nbt)) for nbt in self.not_nbts)

        if self.scores is not None:
            yield 'scores', str(self.scores)

        if self.advancements is not None:
            yield 'advancements', str(self.advancements)

        if self.sort is not None:
            yield 'sort', str(self.sort)

        if self.limit is not None:
            yield 'limit', str(self.limit)

    @property
    @abc.abstractmethod
    def base(self) -> SelectorBase:
        """ Return the base selector type as a `SelectorBase`. For example, the `s` in `@s`. """



import unittest

from pymcutil.selector.advancement_set import AdvancementSet


class AdvancementSetTestCase(unittest.TestCase):
    # TODO more tests

    def test(self):
        self.assertEqual(
            '{foo=true,bar=false,custom:something={criterion=true}}',
            str(AdvancementSet(**{'foo': True, 'bar': False, 'custom:something': {'criterion': True}})))



import abc


class Siftable(abc.ABC):
    DEFAULT_SENTINEL = object()

    @classmethod
    def sift(cls, obj, default=DEFAULT_SENTINEL):
        # If a value was not supplied, try the default.
        if obj is None:
            if default is cls.DEFAULT_SENTINEL:
                raise ValueError('Cannot create {} from nothing, with no default'.format(cls.__name__))
            obj = default

        # If the value explicitly defaulted to null, just return the same.
        if obj is None:
            return None

        # If the value is already an instance of this same class, just return it as is.
        if isinstance(obj, cls):
            return obj

        # Finally, trying converting the value into an instance of this class.
        converted = cls._from_other(obj)
        if converted is not None:
            return converted

        raise ValueError('Cannot convert {} to {}'.format(obj.__class__.__name__, cls.__name__))

    @classmethod
    @abc.abstractmethod
    def _from_other(cls, obj):
        """ Return an instance of `cls` constructed from the non-`cls`, non-null generic object `obj`. """


class SequenceSiftable(Siftable):
    @classmethod
    def _from_other(cls, obj):
        return cls(*obj)


class MappingSiftable(Siftable):
    @classmethod
    def _from_other(cls, obj):
        return cls(**obj)



from .._execute_store_command import ExecuteStoreCommand

CMD = 'result'


class ExecuteStoreResultCommand(ExecuteStoreCommand):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute store result <name: entity> <objective: string> -> execute

    http://minecraft.gamepedia.com/Commands#execute
    """

    def params(self):
        yield from super().params()
        yield from (
            CMD, self.target, self.objective,
            (self.subcommand.substr(), None))


result = ExecuteStoreResultCommand



from typing import Iterable, List

from pymcutil.command.command import Command
from pymcutil.resource.abc.resource import Resource


class FunctionResource(Resource):
    def __init__(self, commands: Iterable[Command]):
        self.commands: List[Command] = list(commands)

    @property
    def lines(self) -> Iterable[str]:
        # Yield commands, one per line, followed by one final blank line.
        yield from (str(command) for command in self.commands)
        yield ''

    def write_to(self, path):
        with open(path, 'w') as fp:
            fp.writelines('\n'.join(self.lines))



import typing
from numbers import Real
from typing import Iterable, Union

from pymcutil.util.siftable import Siftable


class Range(Siftable):
    """
    Represents numeric ranges as they appear in selectors:

        level=10 is level 10
        level=10..12 is level 10, 11 or 12
        level=5.. is anything level 5 or above
        level=..15 is anything level 15 or below

    Supports unbounded minimum, maximum, and exact values by (ab)using the builtin `...` (`Ellipsis`) constant:

        Range(1) => 1
        Range(1, ...) => 1..
        Range(..., 2) => ..2
        Range(1, 2) => 1..2

    Any numeric, iterable, or string objects may be implicitly converted in context:

        1 => Range(1)
        (1, ...) => Range(1, ...)
        [..., 2] => Range(..., 2)
        '1..2' => Range(1, 2)

    """

    Generic = Union['Range', Real, Iterable, str]

    @classmethod
    def _from_other(cls, obj):
        if isinstance(obj, Real):
            return cls.from_real(obj)
        elif isinstance(obj, str):
            return cls.from_string(obj)
        elif isinstance(obj, Iterable):
            return cls.from_iterable(obj)

    @classmethod
    def from_real(cls, n: Real):
        """ Convert real numbers as is_exact values. """
        return cls(n)

    @classmethod
    def from_iterable(cls, i: Iterable):
        """ Convert tuples and lists. """
        return cls(*i)

    @classmethod
    def from_string(cls, s: str):
        """ Convert strings from their natural form. """
        # TODO use regex
        parts = s.split('..')
        min_str = parts[0]
        max_str = parts[1] if len(parts) > 1 else None
        min_ = float(min_str) if min_str else ...
        max_ = float(max_str) if max_str else (... if max_str == '' else None)
        return cls(min_, max_)

    def __init__(self, min_: Real, max_: Real = None):
        self._min: Real = min_
        self._max: Real = max_ if max_ is not None else min_  # Range(1) => Range(1, 1)

    def __str__(self):
        min_str = '{:g}'.format(self.min) if isinstance(self.min, Real) else ''
        if self.is_exact:
            return min_str
        max_str = '{:g}'.format(self.max) if isinstance(self.max, Real) else ''
        return '{}..{}'.format(min_str, max_str)

    def __eq__(self, other):
        return isinstance(other, Range) \
               and isinstance(other, self.__class__) \
               and (self.min == other.min) \
               and (self.max == other.max)

    @property
    def min(self) -> typing.Union[Real, None]:
        return self._min

    @property
    def max(self) -> typing.Union[Real, None]:
        return self._max

    @property
    def is_exact(self) -> bool:
        return isinstance(self.min, Real) and (self.min == self.max)



import abc
import collections
import typing


class ResourceReference(abc.ABC, collections.Hashable):
    @property
    @abc.abstractmethod
    def params(self) -> typing.Iterable:
        """ Yield parameters to be used when generating the resource. """



from pymcutil.resource.resource_manager.abc.resource_manager import ResourceManager
from pymcutil.resource.resource_manager.errors import ReferenceNotRegisteredError, ResourceDidNotGenerateError


class StandardResourceManager(ResourceManager):
    def __init__(self, database, registry):
        self._database = database

        self._generator_map = {}
        self._database_map = {}

        self._generating = set()

        for item in registry:
            self.register(*item)

    def _get_generator(self, reference):
        try:
            return self._generator_map[type(reference)]
        except KeyError as e:
            raise ReferenceNotRegisteredError(reference) from e

    def _get_database(self, reference):
        return self._database_map.get(type(reference), self._database)

    def register(self, kind, generator, database=None):
        self._generator_map[kind] = generator
        self._database_map[kind] = database or self._database

    def get(self, reference):
        generator = self._get_generator(reference)
        database = self._get_database(reference)
        location = generator.location(self, reference)
        resource = database.get(location)
        return resource

    def generate(self, reference):
        self.locate(reference)
        resource = self.get(reference)
        if not resource:
            raise ResourceDidNotGenerateError(reference)
        return resource

    def locate(self, reference):
        located_resource = self.get(reference)

        if located_resource:
            location = located_resource.location

        else:
            generator = self._get_generator(reference)
            location = generator.location(self, reference)

            # Avoid deadlocks due to circular references.
            if reference not in self._generating:
                self._generating.add(reference)

                database = self._get_database(reference)
                resource = generator.generate(self, reference)
                database.put(location, resource)

                self._generating.remove(reference)

        return location

    def name(self, reference):
        return self.locate(reference).name

    def path(self, reference):
        return self.locate(reference).path



from pymcutil.util.argument_mapping import ArgumentMapping


class AdvancementSet(ArgumentMapping):
    def __setitem__(self, key, value):
        if not isinstance(value, bool):
            value = ArgumentMapping.sift(value, None)
            if value is None:
                raise ValueError('Value {} is neither a bool nor another mapping'.format(value))
        super().__setitem__(key, value)




import unittest

from pymcutil.command import commands
from pymcutil.selector import selectors


class ExecuteAsCommandTestCase(unittest.TestCase):
    def test_without_data_tag(self):
        cmd = commands.execute.as_(
            target=selectors.SELF,
            subcommand=commands.execute.then(commands.say('hello')))
        self.assertEqual(str(cmd), 'execute as @s then say hello')



from pymcutil.data_tag import DataTag
from pymcutil.resource.abc.nbt_resource import NBTResource


class StructureResource(NBTResource):
    def __init__(self, nbt):
        self._nbt = nbt

    @property
    def nbt(self) -> DataTag:
        return self._nbt



from typing import Mapping, Union

from pymcutil.command.target import Target
from pymcutil.data_tag.compound_data_tag import CompoundDataTag
from ..scoreboard_players_tag_command import ScoreboardPlayersTagCommand

CMD = 'add'


class ScoreboardPlayersTagAddCommand(ScoreboardPlayersTagCommand):
    """
    An objective model of one of Minecraft's `scoreboard` subcommands:

        scoreboard players tag <player> add <tagName> [dataTag]

    http://minecraft.gamepedia.com/Scoreboard#Tags_commands
    """

    def __init__(self, target: Target, tag: str, data_tag: Mapping = None):
        super().__init__(target)
        self.tag: str = tag
        self.data_tag: Union[CompoundDataTag, None] = CompoundDataTag.sift(data_tag, None)

    def params(self):
        yield from super().params()
        yield from (
            CMD, self.tag,
            (self.data_tag, CompoundDataTag()))


add = ScoreboardPlayersTagAddCommand



import unittest

from pymcutil.block.block_entity import BlockEntity
from pymcutil.block.block_state import BlockState
from pymcutil.data_tag.compound_data_tag import CompoundDataTag
from pymcutil.data_tag.string_data_tag import StringDataTag


class BlockEntityTestCase(unittest.TestCase):
    def test_with_id(self):
        be = BlockEntity('chest')
        self.assertEqual(be.block_id, 'chest')

    def test_with_id_and_block_state(self):
        be = BlockEntity('chest', BlockState(facing='north'))
        self.assertEqual(be.block_id, 'chest')
        self.assertEqual(be.block_state['facing'], 'north')

    def test_with_id_and_block_state_and_data_tag(self):
        be = BlockEntity('chest', BlockState(facing='north'), CompoundDataTag(lock='password'))
        self.assertEqual(be.block_id, 'chest')
        self.assertEqual(be.block_state['facing'], 'north')
        self.assertEqual(be.data_tag['lock'], StringDataTag('password'))

    def test_data_tag_from_dict(self):
        be = BlockEntity('chest', data_tag=dict(lock='password'))
        self.assertEqual(be.data_tag['lock'], StringDataTag('password'))

    def test_data_tag_from_braces(self):
        be = BlockEntity('chest', data_tag={'Lock': 'password'})
        self.assertEqual(be.data_tag['Lock'], StringDataTag('password'))

    def test_tag(self):
        be1 = BlockEntity('chest')
        be2 = be1.tag({'Lock': 'password'})
        self.assertEqual(be1.block_id, 'chest')
        self.assertEqual(be1.data_tag, CompoundDataTag())
        self.assertEqual(be2.block_id, 'chest')
        self.assertEqual(be2.data_tag, CompoundDataTag({'Lock': 'password'}))

    def test_tag_overwrite(self):
        be1 = BlockEntity('chest', data_tag=CompoundDataTag({'Lock': 'password'}))
        be2 = be1.tag({'Lock': 'guessme'})
        self.assertEqual(be1.block_id, 'chest')
        self.assertEqual(be1.data_tag, CompoundDataTag({'Lock': 'password'}))
        self.assertEqual(be2.block_id, 'chest')
        self.assertEqual(be2.data_tag, CompoundDataTag({'Lock': 'guessme'}))

    def test_tag_more(self):
        be1 = BlockEntity('chest', data_tag=CompoundDataTag({'Lock': 'password'}))
        be2 = be1.tag({'CustomName': 'Materials'})
        self.assertEqual(be1.block_id, 'chest')
        self.assertEqual(be1.data_tag, CompoundDataTag({'Lock': 'password'}))
        self.assertEqual(be2.block_id, 'chest')
        self.assertEqual(be2.data_tag, CompoundDataTag({'Lock': 'password', 'CustomName': 'Materials'}))



import logging
import os
from typing import Dict, Iterable, Union

from pymcutil.resource.abc.resource import Resource
from pymcutil.resource.located_resource.located_resource import LocatedResource
from pymcutil.resource.resource_database.abc.resource_database import ResourceDatabase
from pymcutil.resource.resource_location.abc.resource_location import ResourceLocation

log = logging.getLogger(__name__)


class ForgetfulResourceDatabase(ResourceDatabase):
    """
    Never reads, only writes. Relies entirely on an in-memory store for retrieval. Generated resources are written
    to file regardless of whether they already exist.

    Given that it does nothing with existing resources, it's limited to smaller systems and production builds. Better
    to use a more advanced `ResourceDatabase` implementation for anything even remotely complex.
    """

    def __init__(self, data_path: str):
        self.data_path: str = data_path
        self.data: Dict[ResourceLocation, LocatedResource] = {}

    def get(self, location: ResourceLocation) -> Union[LocatedResource, None]:
        result = self.data.get(location, None)
        log.debug('GOT {} from {}'.format(type(result.resource).__name__ if result else None, location.path))
        return result

    def put(self, location: ResourceLocation, resource: Resource) -> LocatedResource:
        log.debug('PUT {} at {}'.format(type(resource).__name__, location.path))

        resource_path = os.path.join(self.data_path, location.path)
        os.makedirs(os.path.dirname(resource_path), exist_ok=True)
        resource.write_to(resource_path)

        self.data[location] = LocatedResource(location, resource)
        return self.get(location)

    def all(self) -> Iterable[LocatedResource]:
        yield from self.data.values()



from pymcutil.util.argument_mapping import ArgumentMapping
from pymcutil.selector.range import Range


class ScoreSet(ArgumentMapping):
    def __setitem__(self, key, value):
        value = Range.sift(value, None)
        if value is None:
            raise ValueError('Value {} cannot be converted to a valid range'.format(value))
        super().__setitem__(key, value)



import unittest

from pymcutil.command import commands
from pymcutil.selector import selectors


class EffectTakeCommandTestCase(unittest.TestCase):
    def test_without_any_optionals(self):
        cmd = commands.effect.take(
            target=selectors.SELF)
        self.assertEqual(str(cmd), 'effect take @s')

    def test_up_to_effect(self):
        cmd = commands.effect.take(
            target=selectors.SELF,
            effect='strength')
        self.assertEqual(str(cmd), 'effect take @s strength')



from pymcutil.command.target import Target
from ..scoreboard_players_command import ScoreboardPlayersCommand

CMD = 'tag'


class ScoreboardPlayersTagCommand(ScoreboardPlayersCommand):
    """
    An objective model of one of Minecraft's `scoreboard` subcommands:

        scoreboard players tag <player> <add|remove|list> ...

    http://minecraft.gamepedia.com/Scoreboard#Tags_commands

    *This command requires a sub-command to be useful.*
    """

    def __init__(self, target: Target):
        self.target: Target = target

    def params(self):
        yield from super().params()
        yield from (CMD, self.target)



from pymcutil.command.command import Command
from pymcutil.command.target import Target
from pymcutil.selector import selectors

CMD = 'function'


class FunctionCommand(Command):
    """
    An objective model of Minecraft's `function` command:

        function <function> [if|unless] [selector]

    http://minecraft.gamepedia.com/Commands#function
    """

    def __init__(self, function: str, mode: str = None, mode_target: Target = None):
        self.function: str = function
        self.mode: str = mode
        self.mode_target: Target = mode_target

    def params(self):
        yield from (CMD, self.function, (self.mode, 'if'), (self.mode_target, selectors.SELF))


function = FunctionCommand



import logging.config

DEFAULT_LEVEL = logging.WARNING
DEFAULT_FMT = '%(asctime)s | %(levelname)-8s | %(message)s'


def install(level=DEFAULT_LEVEL, fmt=DEFAULT_FMT):
    logging.basicConfig(level=level, format=fmt)

    try:
        import sys
        import colorlog

        formatter = colorlog.ColoredFormatter(
            fmt='%(log_color)s' + fmt + '%(reset)s',
            log_colors={
                'DEBUG': 'blue',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'white,bg_red',
            }
        )

        handler = colorlog.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)

        logging.root.handlers.clear()
        logging.root.addHandler(handler)

    except:
        pass



from typing import Any

from pymcutil.resource.abc.json_resource import JsonResource


class RecipeResource(JsonResource):
    def __init__(self, obj: dict):
        self.obj: dict = obj

    @property
    def serialized(self) -> Any:
        return self.obj



import re

from pymcutil.data_tag import DataTag


class StringDataTag(DataTag):
    """ A string data tag, also known as TAG_String. """

    PATTERN = re.compile('^[a-zA-Z0-9._+\-]+$')

    def __init__(self, value=''):
        self._value: str = None
        self.value = value

    def __eq__(self, other):
        return isinstance(other, StringDataTag) and self.value.__eq__(other.value)

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value):
        value = str(value)
        self._value = value

    def to_str(self) -> str:
        # If the string contains syntax-breaking characters, it needs to be surrounded with quotes and any existing
        # quotes need to be escaped.
        if not self.PATTERN.match(self.value):
            return ''.join(('"', self.value.replace('"', '\\"'), '"'))
        return self.value



from pymcutil.resource.resource_reference.function_resource_reference import FunctionResourceReference


class StringSplitFunctionResourceReference(FunctionResourceReference):
    def __init__(self, message: str, indent: int):
        super().__init__(message, indent)


string_split = StringSplitFunctionResourceReference



from ..scoreboard_players_tag_command import ScoreboardPlayersTagCommand

CMD = 'list'


class ScoreboardPlayersTagListCommand(ScoreboardPlayersTagCommand):
    """
    An objective model of one of Minecraft's `scoreboard` subcommands:

        scoreboard players tag <player> list

    http://minecraft.gamepedia.com/Scoreboard#Tags_commands
    """

    def params(self):
        yield from super().params()
        yield CMD


list = ScoreboardPlayersTagListCommand



from typing import Union

from pymcutil.command.target import Target
from .._execute_command import ExecuteCommand


class ExecuteConditionalEntityMixin(object):
    def __init__(self, target: Target, subcommand: ExecuteCommand = None):
        self.target: Target = target
        self.subcommand: Union[ExecuteCommand, None] = subcommand



from pymcutil.selector.abc.selector import Selector
from pymcutil.symbols import selector_bases
from pymcutil.symbols.selector_bases.selector_bases import SelectorBase


class RandomSelector(Selector):
    @property
    def base(self) -> SelectorBase:
        return selector_bases.random


random = RandomSelector

RANDOM = RandomSelector()



import unittest

from pymcutil.block.block import Block
from pymcutil.block.block_entity import BlockEntity
from pymcutil.command import commands
from pymcutil.position.position import ZERO_OFFSET


class SetblockCommandTestCase(unittest.TestCase):
    def test_block_with_id(self):
        cmd = commands.setblock(
            position=ZERO_OFFSET,
            block='dropper')
        self.assertEqual(str(cmd), 'setblock ~ ~ ~ dropper')

    def test_block_with_id_and_state(self):
        cmd = commands.setblock(
            position=ZERO_OFFSET,
            block=Block(
                block_id='dropper',
                block_state=dict(facing='up')))
        self.assertEqual(str(cmd), 'setblock ~ ~ ~ dropper[facing=up]')

    def test_block_with_id_and_state_and_data_tag(self):
        cmd = commands.setblock(
            position=ZERO_OFFSET,
            block=BlockEntity(
                block_id='dropper',
                block_state=dict(facing='up'),
                data_tag={'Lock': 'password'}))
        self.assertEqual(str(cmd), 'setblock ~ ~ ~ dropper[facing=up]{Lock:password}')

    def test_old_block_handling(self):
        cmd = commands.setblock(
            position=ZERO_OFFSET,
            block='dropper',
            old_block_handling='destroy')
        self.assertEqual(str(cmd), 'setblock ~ ~ ~ dropper destroy')



from pymcutil.data_tag import DataTag


def data_taggify(obj):
    # Note that we are using local imports due to circular dependencies.
    if isinstance(obj, DataTag):
        return obj
    elif isinstance(obj, bool):
        from pymcutil.data_tag.byte_data_tag import ByteDataTag
        return ByteDataTag(obj)
    elif isinstance(obj, int):
        from pymcutil.data_tag.int_data_tag import IntDataTag
        return IntDataTag(obj)
    elif isinstance(obj, float):
        from pymcutil.data_tag.double_data_tag import DoubleDataTag
        return DoubleDataTag(obj)
    elif isinstance(obj, str):
        from pymcutil.data_tag.string_data_tag import StringDataTag
        return StringDataTag(obj)
    elif isinstance(obj, list):
        from pymcutil.data_tag.list_data_tag import ListDataTag
        return ListDataTag(obj)
    elif isinstance(obj, dict):
        from pymcutil.data_tag.compound_data_tag import CompoundDataTag
        return CompoundDataTag(obj)
    else:
        raise TypeError(f'Object cannot be converted into a data tag: {obj}')



import unittest

from pymcutil.selector.range import Range


class RangeTestCase(unittest.TestCase):
    def test_undefined(self):
        self.assertEqual(
            '..',
            str(Range(...)))

    def test_exact_int(self):
        self.assertEqual(
            '1',
            str(Range(1)))

    def test_at_least_int(self):
        self.assertEqual(
            '1..',
            str(Range(1, ...)))

    def test_at_most_int(self):
        self.assertEqual(
            '..2',
            str(Range(..., 2)))

    def test_from_to_int(self):
        self.assertEqual(
            '1..2',
            str(Range(1, 2)))

    def test_exact_float(self):
        self.assertEqual(
            '1.234',
            str(Range(1.234)))

    def test_at_least_float(self):
        self.assertEqual(
            '1.234..',
            str(Range(1.234, ...)))

    def test_at_most_float(self):
        self.assertEqual(
            '..2.345',
            str(Range(..., 2.345)))

    def test_from_to_float(self):
        self.assertEqual(
            '1.234..2.345',
            str(Range(1.234, 2.345)))

    def test_zero_conversion(self):
        self.assertEqual(
            Range(0),
            Range.sift(0))

    def test_int_conversion(self):
        self.assertEqual(
            Range(1),
            Range.sift(1))

    def test_float_conversion(self):
        self.assertEqual(
            Range(1.234),
            Range.sift(1.234))

    def test_tuple_conversion_at_least(self):
        self.assertEqual(
            Range(1, ...),
            Range.sift((1, ...)))

    def test_tuple_conversion_at_most(self):
        self.assertEqual(
            Range(..., 2),
            Range.sift((..., 2)))

    def test_tuple_conversion_from_to(self):
        self.assertEqual(
            Range(1, 2),
            Range.sift((1, 2)))

    def test_list_conversion_at_least(self):
        self.assertEqual(
            Range(1, ...),
            Range.sift([1, ...]))

    def test_list_conversion_at_most(self):
        self.assertEqual(
            Range(..., 2),
            Range.sift([..., 2]))

    def test_list_conversion_from_to(self):
        self.assertEqual(
            Range(1, 2),
            Range.sift([1, 2]))

    def test_string_conversion_exact_int(self):
        self.assertEqual(
            Range(1),
            Range.sift('1'))

    def test_string_conversion_at_least_int(self):
        self.assertEqual(
            Range(1, ...),
            Range.sift('1..'))

    def test_string_conversion_at_most_int(self):
        self.assertEqual(
            Range(..., 2),
            Range.sift('..2'))

    def test_string_conversion_from_to_int(self):
        self.assertEqual(
            Range(1, 2),
            Range.sift('1..2'))

    def test_string_conversion_exact_float(self):
        self.assertEqual(
            Range(1.234),
            Range.sift('1.234'))

    def test_string_conversion_at_least_float(self):
        self.assertEqual(
            Range(1.234, ...),
            Range.sift('1.234..'))

    def test_string_conversion_at_most_float(self):
        self.assertEqual(
            Range(..., 2.345),
            Range.sift('..2.345'))

    def test_string_conversion_from_to_float(self):
        self.assertEqual(
            Range(1.234, 2.345),
            Range.sift('1.234..2.345'))



from .._execute_conditional_blocks_mixin import ExecuteConditionalBlocksMixin
from .._execute_if_command import ExecuteIfCommand

CMD = 'blocks'


class ExecuteIfBlocksCommand(ExecuteIfCommand, ExecuteConditionalBlocksMixin):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute if blocks <start: x y z> <end: x y z> <destination: x y z> ...

    http://minecraft.gamepedia.com/Commands#execute

    *This command requires a sub-command to be useful.*
    """

    def params(self):
        yield from super().params()
        yield from (CMD, self.start, self.end, self.destination)



from typing import Union

from pymcutil.block.block import Block
from pymcutil.command.command import Command
from pymcutil.position.position import Position

CMD = 'setblock'


class SetblockCommand(Command):
    """
    An objective model of Minecraft's `setblock` command:

        setblock <x y z> <block> [oldBlockHandling]

    http://minecraft.gamepedia.com/Commands#setblock
    """

    def __init__(self, position: Position.Generic, block: Block.Generic, old_block_handling: str = None):
        self.position: Position = Position.sift(position)
        self.block: Block = Block.sift(block)
        self.old_block_handling: Union[str, None] = old_block_handling

    def params(self):
        yield from (
            CMD, self.position, self.block,
            (self.old_block_handling, 'replace'))


setblock = SetblockCommand



import unittest

from pymcutil.selector.score_set import ScoreSet


class ScoreSetTestCase(unittest.TestCase):
    # TODO more tests

    def test(self):
        self.assertEqual(
            '{foo=1,bar=1..5}',
            str(ScoreSet(foo=1, bar=(1, 5))))



import unittest

from pymcutil.data_tag.byte_data_tag import ByteDataTag
from pymcutil.data_tag.compound_data_tag import CompoundDataTag
from pymcutil.item.item import Item


class ItemTestCase(unittest.TestCase):
    def test_with_id(self):
        i = Item('diamond')
        self.assertEqual(i.item_id, 'diamond')

    def test_with_id_and_data_tag(self):
        i = Item('diamond', data_tag=CompoundDataTag(count=ByteDataTag(64)))
        self.assertEqual(i.item_id, 'diamond')
        self.assertEqual(i.data_tag['count'], ByteDataTag(64))

    def test_data_tag_from_dict(self):
        i = Item('diamond', data_tag=dict(count=ByteDataTag(64)))
        self.assertEqual(i.data_tag['count'], ByteDataTag(64))

    def test_data_tag_from_braces(self):
        i = Item('diamond', data_tag={'Count': ByteDataTag(64)})
        self.assertEqual(i.data_tag['Count'], ByteDataTag(64))

    def test_tag(self):
        i1 = Item('diamond')
        i2 = i1.tag({'Count': ByteDataTag(64)})
        self.assertEqual(i1.item_id, 'diamond')
        self.assertEqual(i1.data_tag, CompoundDataTag())
        self.assertEqual(i2.item_id, 'diamond')
        self.assertEqual(i2.data_tag, CompoundDataTag({'Count': ByteDataTag(64)}))

    def test_tag_overwrite(self):
        i1 = Item('diamond', data_tag=CompoundDataTag({'Count': ByteDataTag(64)}))
        i2 = i1.tag({'Count': ByteDataTag(32)})
        self.assertEqual(i1.item_id, 'diamond')
        self.assertEqual(i1.data_tag, CompoundDataTag({'Count': ByteDataTag(64)}))
        self.assertEqual(i2.item_id, 'diamond')
        self.assertEqual(i2.data_tag, CompoundDataTag({'Count': ByteDataTag(32)}))

    def test_tag_more(self):
        i1 = Item('diamond', data_tag=CompoundDataTag({'Count': ByteDataTag(64)}))
        i2 = i1.tag({'Slot': ByteDataTag(8)})
        self.assertEqual(i1.item_id, 'diamond')
        self.assertEqual(i1.data_tag, CompoundDataTag({'Count': ByteDataTag(64)}))
        self.assertEqual(i2.item_id, 'diamond')
        self.assertEqual(i2.data_tag, CompoundDataTag({'Count': ByteDataTag(64), 'Slot': ByteDataTag(8)}))



from typing import Union

from pymcutil.command.target import Target
from .._execute_command import ExecuteCommand

CMD = 'store'


class ExecuteStoreCommand(ExecuteCommand):
    """
    An objective model of one of Minecraft's `execute` subcommands:

        execute store ...

    http://minecraft.gamepedia.com/Commands#execute

    *This command requires a sub-command to be useful.*
    """

    def __init__(self, target: Target, objective: str, subcommand: ExecuteCommand = None):
        self.target: Target = target
        self.objective: str = objective
        self.subcommand: Union[ExecuteCommand, None] = subcommand

    def params(self):
        yield from super().params()
        yield CMD



from ..scoreboard_command import ScoreboardCommand

CMD = 'players'


class ScoreboardPlayersCommand(ScoreboardCommand):
    """
    An objective model of one of Minecraft's `scoreboard` subcommands:

        scoreboard players <list|set|add|remove|reset|enable|test|operation|tag> ...

    http://minecraft.gamepedia.com/Scoreboard#Players_commands

    *This command requires a sub-command to be useful.*
    """

    def params(self):
        yield from super().params()
        yield CMD



import unittest

from pymcutil.block.block_state import BlockState
from pymcutil.position.position import Position


class SiftableTestCase(unittest.TestCase):
    def test_sift_position_from_tuple(self):
        self.assertEqual(Position.sift((1, 2, 3)), Position(1, 2, 3))

    def test_sift_position_from_list(self):
        self.assertEqual(Position.sift([1, 2, 3]), Position(1, 2, 3))

    def test_sift_position_from_dict(self):
        self.assertEqual(Position.sift({1: 'one', 2: 'two', 3: 'three'}), Position(1, 2, 3))

    def test_sift_position_from_position(self):
        self.assertEqual(Position.sift(Position(1, 2, 3)), Position(1, 2, 3))

    def test_sift_block_state_from_dict(self):
        self.assertEqual(BlockState.sift({'a': 'alpha', 'b': 'beta'}), BlockState(a='alpha', b='beta'))

    def test_sift_block_state_from_block_state(self):
        self.assertEqual(BlockState.sift(BlockState(a='alpha', b='beta')), BlockState(a='alpha', b='beta'))

    def test_siftable_default(self):
        self.assertEqual(Position.sift(None, (1, 2, 3)), Position(1, 2, 3))

    def test_siftable_empty_default(self):
        self.assertIsNone(Position.sift(None, None))

    def test_sift_value_error(self):
        with self.assertRaises(ValueError):
            Position.sift(None)

    def test_sift_type_error(self):
        with self.assertRaises(TypeError):
            Position.sift(42)



import unittest

from pymcutil.block.block_state import BlockState


class BlockStateTestCase(unittest.TestCase):
    def test_eq(self):
        bs = BlockState(color='blue')
        self.assertEqual(bs, BlockState(color='blue'))

    def test_getitem(self):
        bs = BlockState(half='bottom', variant='quartz')
        self.assertEqual(bs['half'], 'bottom')
        self.assertEqual(bs['variant'], 'quartz')

    def test_len_0(self):
        bs0 = BlockState()
        self.assertEqual(len(bs0), 0)

    def test_len_1(self):
        bs1 = BlockState(color='blue')
        self.assertEqual(len(bs1), 1)

    def test_len_2(self):
        bs2 = BlockState(half='bottom', variant='quartz')
        self.assertEqual(len(bs2), 2)

    def test_delitem(self):
        bs = BlockState(half='bottom', variant='quartz')
        self.assertEqual(bs['variant'], 'quartz')
        del bs['variant']
        with self.assertRaises(KeyError):
            shouldfail = bs['variant']

    def test_setitem(self):
        bs = BlockState()
        bs['color'] = 'blue'
        self.assertEqual(bs['color'], 'blue')

    def test_setitem_bad_key_str(self):
        bs = BlockState()
        with self.assertRaises(KeyError):
            bs['bad key'] = 'goodvalue'

    def test_setitem_bad_key_obj(self):
        bs = BlockState()
        with self.assertRaises(KeyError):
            bs[{'bad', 'obj'}] = 'goodvalue'

    def test_setitem_bad_value_str(self):
        bs = BlockState()
        with self.assertRaises(ValueError):
            bs['goodkey'] = 'bad value'

    def test_setitem_bad_value_obj(self):
        bs = BlockState()
        with self.assertRaises(ValueError):
            bs['goodkey'] = {'bad', 'obj'}

    def test_set_then_get(self):
        bs = BlockState()
        bs['color'] = 'blue'
        self.assertEqual(bs['color'], 'blue')

    def test_overwrite(self):
        bs = BlockState(color='blue')
        self.assertEqual(bs['color'], 'blue')
        bs['color'] = 'red'
        self.assertEqual(bs['color'], 'red')

    def test_bool_false(self):
        bs = BlockState(powered=False)
        self.assertEqual(bs['powered'], False)

    def test_bool_true(self):
        bs = BlockState(powered=True)
        self.assertEqual(bs['powered'], True)

    def test_int_value(self):
        bs = BlockState(age=4)
        self.assertEqual(bs['age'], 4)

    def test_str_default(self):
        bs = BlockState()
        self.assertEqual(str(bs), 'default')

    def test_str_with_str_values(self):
        bs = BlockState(half='bottom', variant='quartz')
        self.assertEqual(
            set(str(bs).split(',')),
            set('half=bottom,variant=quartz'.split(',')))

    def test_str_with_all_values(self):
        bs = BlockState(powered=True, age=7, color='green')
        self.assertEqual(
            set(str(bs).split(',')),
            set('powered=true,age=7,color=green'.split(',')))



from typing import Union

from pymcutil.position.position import Position
from .._execute_command import ExecuteCommand


class ExecuteConditionalBlocksMixin(object):
    def __init__(
            self, start: Position.Generic, end: Position.Generic, destination: Position.Generic,
            subcommand: ExecuteCommand = None):
        self.start: Position = Position.sift(start)
        self.end: Position = Position.sift(end)
        self.destination: Position = Position.sift(destination)
        self.subcommand: Union[ExecuteCommand, None] = subcommand



class RC(float):
    """ A float wrapper for dealing with relative coordinates. """

    def __repr__(self):
        return '{name}({params})'.format(
            name=self.__class__.__name__,
            params=super().__repr__())

    def __str__(self):
        return '~' + (super().__str__() if self else '')

    def __eq__(self, other):
        return (isinstance(other, RC)) and super().__eq__(other)

    def __add__(self, other):
        try:
            return RC(super().__add__(other))
        except Exception as e:
            raise TypeError('{self!r} cannot add with {other!r}'.format(self=self, other=other)) from e

    def __sub__(self, other):
        try:
            return RC(super().__sub__(other))
        except Exception as e:
            raise TypeError('{self!r} cannot sub with {other!r}'.format(self=self, other=other)) from e

    def __mul__(self, other):
        try:
            return RC(super().__mul__(other))
        except Exception as e:
            raise TypeError('{self!r} cannot mul with {other!r}'.format(self=self, other=other)) from e

    def __truediv__(self, other):
        try:
            return RC(super().__truediv__(other))
        except Exception as e:
            raise TypeError('{self!r} cannot truediv with {other!r}'.format(self=self, other=other)) from e

    def __floordiv__(self, other):
        try:
            return RC(super().__floordiv__(other))
        except Exception as e:
            raise TypeError('{self!r} cannot floordiv with {other!r}'.format(self=self, other=other)) from e

    def __iadd__(self, other):
        return self.__add__(other)

    def __isub__(self, other):
        return self.__sub__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __itruediv__(self, other):
        return self.__truediv__(other)

    def __ifloordiv__(self, other):
        return self.__floordiv__(other)

    # Keep `radd` and `rmul` disabled so the builtin float versions are called instead whenever a relative coordinate
    # appears on the right-hand side of the operation. This causes the resulting type of `add` and `mul` to depend on
    # whichever comes first.

    # def __radd__(self, other):
    #     return self.__add__(other)
    #
    # def __rmul__(self, other):
    #     return self.__mul__(other)

    def __neg__(self):
        return self.__mul__(-1)

    def __format__(self, format_spec):
        return '~' + (super().__format__(format_spec) if self else '')


def rel(value):
    """ Shortcut method for creating individual relative coordinates. """
    return RC(value)



import unittest

from pymcutil.data_tag.float_data_tag import FloatDataTag


class FloatDataTagTestCase(unittest.TestCase):
    def test_default(self):
        dt = FloatDataTag()
        self.assertEqual(dt.value, 0.0)
        self.assertEqual(dt.to_str(), '0.0f')

    def test_positive(self):
        dt = FloatDataTag(123.5)
        self.assertEqual(dt.value, 123.5)
        self.assertEqual(dt.to_str(), '123.5f')

    def test_negative(self):
        dt = FloatDataTag(-123.5)
        self.assertEqual(dt.value, -123.5)
        self.assertEqual(dt.to_str(), '-123.5f')

    def test_int_conversion(self):
        dt = FloatDataTag(123)
        self.assertEqual(dt.value, 123.0)
        self.assertEqual(dt.to_str(), '123.0f')

    def test_equality(self):
        dt1 = FloatDataTag(123.5)
        dt2 = FloatDataTag(123.5)
        self.assertEqual(dt1, dt2)



import unittest

from pymcutil.data_tag.list_data_tag import ListDataTag
from pymcutil.data_tag.string_data_tag import StringDataTag


class ListDataTagTestCase(unittest.TestCase):
    def test_get(self):
        dt = ListDataTag(['item'])
        self.assertEqual(dt[0], StringDataTag('item'))
        self.assertEqual(dt.to_str(), '[item]')

    def test_append(self):
        dt = ListDataTag()
        dt.append('item')
        self.assertEqual(dt[0], StringDataTag('item'))
        self.assertEqual(dt.to_str(), '[item]')

    def test_insert(self):
        dt = ListDataTag(['item', 'anotheritem'])
        dt.insert(1, 'middleitem')
        self.assertEqual(dt[1], StringDataTag('middleitem'))
        self.assertEqual(dt.to_str(), '[item,middleitem,anotheritem]')

    def test_overwrite(self):
        dt = ListDataTag(['item'])
        dt[0] = 'anotheritem'
        self.assertEqual(dt[0], StringDataTag('anotheritem'))
        self.assertEqual(dt.to_str(), '[anotheritem]')

    def test_del_only(self):
        dt = ListDataTag(['item'])
        self.assertEqual(dt.to_str(), '[item]')
        del dt[0]
        self.assertEqual(dt.to_str(), '[]')

    def test_del_with_remaining(self):
        dt = ListDataTag(['item', 'anotheritem'])
        self.assertEqual(dt.to_str(), '[item,anotheritem]')
        del dt[0]
        self.assertEqual(dt.to_str(), '[anotheritem]')

    def test_default(self):
        dt = ListDataTag()
        self.assertEqual(dt.to_str(), '[]')

    def test_empty_list(self):
        dt = ListDataTag([])
        self.assertEqual(dt.to_str(), '[]')

    def test_multi_empty_lists(self):
        dt = ListDataTag([], [])
        self.assertEqual(dt.to_str(), '[[],[]]')

    def test_multi_nonempty_lists(self):
        dt = ListDataTag(['first'], ['second'])
        self.assertEqual(dt.to_str(), '[[first],[second]]')

    def test_breaking_string(self):
        dt = ListDataTag(['my:string'])
        self.assertEqual(dt.to_str(), '["my:string"]')

    def test_quoted_string(self):
        dt = ListDataTag(['my"string'])
        self.assertEqual(dt.to_str(), r'["my\"string"]')

    def test_init_with_many_values(self):
        dt = ListDataTag([True, 123456789, 123.5, 'mystring', [], {}])
        self.assertEqual(dt.to_str(), '[1b,123456789,123.5d,mystring,[],{}]')

    def test_append_with_many_values(self):
        dt = ListDataTag()
        dt.append(True)
        dt.append(123456789)
        dt.append(123.5)
        dt.append('mystring')
        dt.append([])
        dt.append({})
        self.assertEqual(dt.to_str(), '[1b,123456789,123.5d,mystring,[],{}]')

    def test_addition_with_many_values(self):
        dt = ListDataTag()
        dt += [True, 123456789, 123.5, 'mystring', [], {}]
        self.assertEqual(dt.to_str(), '[1b,123456789,123.5d,mystring,[],{}]')

    def test_deep_conversion(self):
        dt = ListDataTag([[[[[True, 123456789, 123.5, 'mystring', [], {}]]]]])
        self.assertEqual(dt.to_str(), '[[[[[1b,123456789,123.5d,mystring,[],{}]]]]]')



import unittest

from pymcutil.data_tag.string_data_tag import StringDataTag


class StringDataTagTestCase(unittest.TestCase):
    def test_default(self):
        dt = StringDataTag()
        self.assertEqual(dt.value, '')
        self.assertEqual(dt.to_str(), '""')

    def test_empty_string(self):
        dt = StringDataTag('')
        self.assertEqual(dt.value, '')
        self.assertEqual(dt.to_str(), '""')

    def test_simple_string(self):
        dt = StringDataTag('mystring')
        self.assertEqual(dt.value, 'mystring')
        self.assertEqual(dt.to_str(), 'mystring')

    def test_int_conversion(self):
        dt = StringDataTag(123)
        self.assertEqual(dt.value, '123')
        self.assertEqual(dt.to_str(), '123')

    def test_float_conversion(self):
        dt = StringDataTag(12.34)
        self.assertEqual(dt.value, '12.34')
        self.assertEqual(dt.to_str(), '12.34')

    def test_breaking_str(self):
        dt = StringDataTag('my:string')
        self.assertEqual(dt.value, 'my:string')
        self.assertEqual(dt.to_str(), '"my:string"')

    def test_quoted_str(self):
        dt = StringDataTag('"mystring"')
        self.assertEqual(dt.value, '"mystring"')
        self.assertEqual(dt.to_str(), r'"\"mystring\""')

    def test_equality(self):
        dt1 = StringDataTag('hello')
        dt2 = StringDataTag('hello')
        self.assertEqual(dt1, dt2)



from collections import Mapping
from typing import Union

from pymcutil.block.block_state import BlockState
from pymcutil.util.siftable import Siftable


class Block(Siftable):
    """ Represents a Minecraft block. """

    Generic = Union['Block', str]

    def __init__(self, block_id: str, block_state: Mapping = None):
        self.block_id: str = block_id
        self.block_state: BlockState = BlockState.sift(block_state, {})

    def __str__(self):
        return ''.join((
            self.block_id,
            '[{}]'.format(self.block_state) if self.block_state else ''))

    @classmethod
    def _from_other(cls, obj):
        if isinstance(obj, str):
            return Block(block_id=obj)  # TODO parse state and nbt

    def state(self, *args, **params):
        new_block_state = dict(self.block_state)  # copy
        new_block_state.update(*args, **params)
        return Block(block_id=self.block_id, block_state=new_block_state)



import typing

from pymcutil.resource.resource_location.standard_resource_location import StandardResourceLocation


class TagResourceLocation(StandardResourceLocation):
    def __init__(self, namespace: str, components: typing.Iterable[str]):
        super().__init__(namespace=namespace, components=components, subfolder='tags')

    def __str__(self):
        return '#{}'.format(super().__str__())



from collections import Iterable, MutableSequence

from pymcutil.data_tag import DataTag
from pymcutil.data_tag.helper import data_taggify
from pymcutil.util.siftable import SequenceSiftable


class ListDataTag(DataTag, MutableSequence, SequenceSiftable):
    """ A list data tag, also known as TAG_List. """

    Generic = Iterable

    def __init__(self, *args):
        self._value = []

        # If the only thing supplied is an iterable, then append everything from it.
        if len(args) == 1 and isinstance(args[0], Iterable):
            for item in args[0]:
                self.append(item)

        # Otherwise, append everything given.
        else:
            for arg in args:
                self.append(arg)

    def __delitem__(self, index):
        return self._value.__delitem__(index)

    def __setitem__(self, index, value):
        value = data_taggify(value)
        return self._value.__setitem__(index, value)

    def __len__(self):
        return self._value.__len__()

    def __getitem__(self, index):
        return self._value.__getitem__(index)

    def __eq__(self, other):
        return isinstance(other, ListDataTag) and self.value.__eq__(other.value)

    @property
    def value(self) -> list:
        return self._value

    def insert(self, index, value):
        value = data_taggify(value)
        return self._value.insert(index, value)

    def to_str(self) -> str:
        items = (str(item) for item in self)
        return ''.join(('[', ','.join(items), ']'))



import abc


class DataTag(abc.ABC):
    """ A generic data tag, to be used as an interface for all TAG types. """

    def __str__(self):
        return self.to_str()

    def __bytes__(self):
        return self.to_bytes()

    @abc.abstractmethod
    def to_str(self) -> str:
        """ Return a string serialization of the data tag, for example with quotes or a type suffix. """

    def to_bytes(self) -> bytes:
        """ Return a bytes representation of the data tag. """



import unittest

from pymcutil import util


class UtilTestCase(unittest.TestCase):
    def test_default_with_value(self):
        self.assertEqual(util.default('hello'), 'hello')

    def test_default_with_value_and_default(self):
        self.assertEqual(util.default('hello', 'world'), 'hello')

    def test_default_with_value_none_and_default(self):
        self.assertEqual(util.default(None, 'world'), 'world')

    def test_default_with_value_none_and_default_none(self):
        self.assertEqual(util.default(None, None), None)

    def test_defaults(self):
        self.assertEqual(
            util.defaults(dict(a='alpha', b='beta'), b='bravo', c='charlie'),
            dict(a='alpha', b='beta', c='charlie'))

    def test_first_immediate(self):
        self.assertEqual(util.first('a', 'b', 'c'), 'a')

    def test_first_eventually(self):
        self.assertEqual(util.first(None, None, None, 'a', 'b', 'c'), 'a')

    def test_first_none(self):
        self.assertEqual(util.first(None, None, None), None)

    def test_require(self):
        self.assertEqual(util.require(123, 'number'), 123)

    def test_require_error(self):
        with self.assertRaises(ValueError):
            util.require(None, 'number')

    def test_get_logger(self):
        log = util.get_logger([], 'mylist')
        self.assertEqual(log.name, 'list:mylist')



from pymcutil.command.command import Command

CMD = 'scoreboard'


class ScoreboardCommand(Command):
    """
    An objective model of Minecraft's `scoreboard` command:

        scoreboard <objectives|players|teams> ...

    http://minecraft.gamepedia.com/Scoreboard#Command_reference

    *This command requires a sub-command to be useful.*
    """

    def params(self):
        yield CMD



import logging


def default(value, default_=None):
    return default_ if value is None else value


def defaults(original: dict, **kwargs):
    kwargs.update(original)
    return kwargs


def first(*args):
    return next((item for item in args if item is not None), None)


def require(value, name):
    if value is None:
        raise ValueError('Value "{}" is required but missing'.format(name))
    return value


def instance_label(obj: object, label: str = '') -> str:
    return obj.__class__.__name__ + (':{}'.format(label) if label else '')


def get_logger(obj: object, label: str = '') -> logging.Logger:
    return logging.getLogger(instance_label(obj, label))



import unittest

from pymcutil.constants import BYTE_MIN, BYTE_MAX
from pymcutil.data_tag.byte_data_tag import ByteDataTag


class ByteDataTagTestCase(unittest.TestCase):
    def test_default(self):
        dt = ByteDataTag()
        self.assertEqual(dt.value, 0)
        self.assertEqual(dt.to_str(), '0b')

    def test_positive(self):
        dt = ByteDataTag(1)
        self.assertEqual(dt.value, 1)
        self.assertEqual(dt.to_str(), '1b')

    def test_negative(self):
        dt = ByteDataTag(-1)
        self.assertEqual(dt.value, -1)
        self.assertEqual(dt.to_str(), '-1b')

    def test_float_conversion(self):
        dt = ByteDataTag(1.5)
        self.assertEqual(dt.value, 1)
        self.assertEqual(dt.to_str(), '1b')

    def test_min(self):
        dt = ByteDataTag(BYTE_MIN)
        self.assertEqual(dt.value, BYTE_MIN)
        self.assertEqual(dt.to_str(), f'{BYTE_MIN}b')

    def test_max(self):
        dt = ByteDataTag(BYTE_MAX)
        self.assertEqual(dt.value, BYTE_MAX)
        self.assertEqual(dt.to_str(), f'{BYTE_MAX}b')

    def test_too_low(self):
        with self.assertRaises(ValueError):
            ByteDataTag(BYTE_MIN - 1)

    def test_too_high(self):
        with self.assertRaises(ValueError):
            ByteDataTag(BYTE_MAX + 1)

    def test_equality(self):
        dt1 = ByteDataTag(1)
        dt2 = ByteDataTag(1)
        self.assertEqual(dt1, dt2)



from setuptools import setup, find_packages

with open('VERSION.txt') as fp:
    version = fp.read().strip()

with open('README.md') as fp:
    readme = fp.read()

with open('requirements.txt') as fp:
    requirements = fp.read().splitlines()

setup(
    name='pymcutil',
    version=version,
    author='Arcensoth',
    author_email='arcensoth@gmail.com',
    url='https://github.com/Arcensoth/pymcutil',
    license='MIT',
    description='An expressive Minecraft utility library revolving around data manipulation and generation.',
    long_description_content_type='text/markdown',
    long_description=readme,
    packages=find_packages(),
    install_requires=requirements,
    python_requires='>=3.6',
    keywords='minecraft commands library utility datapack data',
    classifiers=(
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
    ))



import abc
import collections
import typing

ResourceName = str
ResourcePath = str


class ResourceLocation(abc.ABC, collections.Hashable):
    @property
    @abc.abstractmethod
    def namespace(self) -> str:
        """ Return the `namespace` in `namespace:path/to/file`. """

    @property
    @abc.abstractmethod
    def components(self) -> typing.Iterable[str]:
        """ Yield the directory components of `path/to/file` in `namespace:path/to/file`. """

    @property
    @abc.abstractmethod
    def trail(self) -> str:
        """ Return the `path/to/file` in `namespace:path/to/file`. """

    @property
    @abc.abstractmethod
    def subfolder(self) -> str:
        """ Return the datapack subfolder where the resource should reside. """

    @property
    @abc.abstractmethod
    def extension(self) -> str:
        """ Return the file extension of the resource. """

    @property
    @abc.abstractmethod
    def name(self) -> ResourceName:
        """ Return the name of the resource, to be used as a reference. """

    @property
    @abc.abstractmethod
    def path(self) -> ResourcePath:
        """ Return the file path of the resource, relative to datapack root. """



from pymcutil.util.symbol import Symbol


class SelectorBase(Symbol):
    pass


all_players = SelectorBase('a')
entities = SelectorBase('e')
player = SelectorBase('p')
random = SelectorBase('r')
self = SelectorBase('s')



from typing import Dict, Iterable, Union

from pymcutil.resource.abc.resource import Resource
from pymcutil.resource.located_resource.located_resource import LocatedResource
from pymcutil.resource.resource_database.abc.resource_database import ResourceDatabase
from pymcutil.resource.resource_location.abc.resource_location import ResourceLocation


class DummyResourceDatabase(ResourceDatabase):
    """ Never reads and never writes. Stores generated resources in-memory and that's about it. """

    def __init__(self):
        self.data: Dict[ResourceLocation, LocatedResource] = {}

    def get(self, location: ResourceLocation) -> Union[LocatedResource, None]:
        return self.data.get(location, None)

    def put(self, location: ResourceLocation, resource: Resource) -> LocatedResource:
        self.data[location] = LocatedResource(location, resource)
        return self.get(location)

    def all(self) -> Iterable[LocatedResource]:
        yield from self.data.values()



from collections import Mapping

from pymcutil.data_tag.compound_data_tag import CompoundDataTag


class Item(object):
    """ Represents a Minecraft item. """

    def __init__(self, item_id: str, data_tag: Mapping = None):
        self.item_id: str = item_id
        self.data_tag: CompoundDataTag = CompoundDataTag.sift(data_tag, {})

    def tag(self, *args, **params):
        new_data_tag = dict(self.data_tag)  # copy
        new_data_tag.update(*args, **params)
        return Item(item_id=self.item_id, data_tag=new_data_tag)



import typing

from pymcutil.resource.resource_reference.abc.resource_reference import ResourceReference


class StandardResourceReference(ResourceReference):
    def __init__(self, *params):
        self._params: typing.Tuple[typing.Any] = tuple(params)  # save a copy

    def __hash__(self):
        return hash(self._params)

    def __eq__(self, other):
        # TODO needs tested
        return isinstance(other, StandardResourceReference) \
               and isinstance(other, self.__class__) \
               and (p1 == p2 for p1, p2 in zip(self.params, other.params))

    @property
    def params(self) -> typing.Iterable[typing.Any]:
        yield from self._params



import unittest

from pymcutil.command import commands
from pymcutil.selector import selectors


class EffectGiveCommandTestCase(unittest.TestCase):
    def test_without_any_optionals(self):
        cmd = commands.effect.give(
            target=selectors.SELF,
            effect='strength')
        self.assertEqual(str(cmd), 'effect give @s strength')

    def test_default_optionals(self):
        cmd = commands.effect.give(
            target=selectors.SELF,
            effect='strength',
            hide_particles=True)
        self.assertEqual(str(cmd), 'effect give @s strength 1000000 0 true')

    def test_up_to_seconds(self):
        cmd = commands.effect.give(
            target=selectors.SELF,
            effect='strength',
            seconds=60)
        self.assertEqual(str(cmd), 'effect give @s strength 60')

    def test_up_to_amplifier(self):
        cmd = commands.effect.give(
            target=selectors.SELF,
            effect='strength',
            seconds=60,
            amplifier=5)
        self.assertEqual(str(cmd), 'effect give @s strength 60 5')

    def test_up_to_hide_particles(self):
        cmd = commands.effect.give(
            target=selectors.SELF,
            effect='strength',
            seconds=60,
            amplifier=5,
            hide_particles=True)
        self.assertEqual(str(cmd), 'effect give @s strength 60 5 true')



from pymcutil.util.common_mapping import CommonMapping


class ArgumentMapping(CommonMapping):
    def __str__(self):
        return ''.join(('{', self.innards, '}'))

    @staticmethod
    def _convert(value):
        if isinstance(value, bool):
            return 'true' if value else 'false'
        return str(value)

    @property
    def innards(self):
        return ','.join(['{}={}'.format(k, self._convert(v)) for k, v in self.items()])



import unittest

from pymcutil.command import commands
from pymcutil.data_tag.compound_data_tag import CompoundDataTag
from pymcutil.selector import selectors


class ScoreboardPlayersTagAddCommandTestCase(unittest.TestCase):
    def test_without_data_tag(self):
        cmd = commands.scoreboard.players.tag.add(
            target=selectors.SELF,
            tag='some_tag')
        self.assertEqual(str(cmd), 'scoreboard players tag @s add some_tag')

    def test_with_data_tag(self):
        cmd = commands.scoreboard.players.tag.add(
            target=selectors.SELF,
            tag='on_ground',
            data_tag=CompoundDataTag({'OnGround': True}))
        self.assertEqual(str(cmd), 'scoreboard players tag @s add on_ground {OnGround:1b}')

    def test_missing_required_param(self):
        with self.assertRaises(ValueError):
            cmd = commands.scoreboard.players.tag.add(
                target=selectors.SELF,
                tag=None)
            str(cmd)
